#!/usr/bin/env python
# encoding: utf-8
#
# bpt.py
#
# Created by José Sánchez-Gallego on 19 Jan 2017.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from packaging.version import parse
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from mpl_toolkits.axes_grid1 import ImageGrid

from marvin.core.exceptions import MarvinDeprecationWarning, MarvinError
from marvin.utils.plot import bind_to_figure


__ALL__ = ('get_snr', 'kewley_sf_nii', 'kewley_sf_sii', 'kewley_sf_oi',
           'kewley_comp_nii', 'kewley_agn_sii', 'kewley_agn_oi',
           'bpt_kewley06')


def get_snr(snr_min, emission_line, default=3):
    """Convenience function to get the minimum SNR for a certain emission line.

    If ``snr_min`` is a dictionary and ``emision_line`` is one of the keys,
    returns that value. If the emission line is not included in the dictionary,
    returns ``default``. If ``snr_min`` is a float, returns that value
    regardless of the ``emission_line``.

    """

    if not isinstance(snr_min, dict):
        return snr_min

    if emission_line in snr_min:
        return snr_min[emission_line]
    else:
        return default


def get_masked(maps, emline, snr=1):
    """Convenience function to get masked arrays without negative values."""

    gflux = maps['emline_gflux_' + emline]
    gflux_masked = gflux.masked

    # Masks spaxels with flux <= 0
    gflux_masked.mask |= (gflux_masked.data <= 0)

    # Masks all spaxels that don't reach the cutoff SNR
    gflux_masked.mask |= gflux.snr < snr
    gflux_masked.mask |= gflux.ivar == 0

    return gflux_masked


def _get_kewley06_axes(use_oi=True):
    """Creates custom axes for displaying Kewley06 plots."""

    fig = plt.figure(None, (8.5, 10))
    fig.clf()

    plt.subplots_adjust(top=0.99, bottom=0.08, hspace=0.01)

    # The axes for the three classification plots
    imgrid_kwargs = {'add_all': True} if parse(matplotlib.__version__) < parse('3.5.0') else {}
    grid_bpt = ImageGrid(fig, 211,
                         nrows_ncols=(1, 3) if use_oi else (1, 2),
                         direction='row',
                         axes_pad=0.1,
                         label_mode='L',
                         share_all=False, **imgrid_kwargs)

    # The axes for the galaxy display
    gal_bpt = ImageGrid(fig, 212, nrows_ncols=(1, 1))

    # Plots the classification boundary lines
    xx_sf_nii = np.linspace(-1.281, 0.045, int(1e4))
    xx_sf_sii = np.linspace(-2, 0.315, int(1e4))
    xx_sf_oi = np.linspace(-2.5, -0.7, int(1e4))

    xx_comp_nii = np.linspace(-2, 0.4, int(1e4))

    xx_agn_sii = np.array([-0.308, 1.0])
    xx_agn_oi = np.array([-1.12, 0.5])

    grid_bpt[0].plot(xx_sf_nii, kewley_sf_nii(xx_sf_nii), 'k--', zorder=90)
    grid_bpt[1].plot(xx_sf_sii, kewley_sf_sii(xx_sf_sii), 'r-', zorder=90)
    if use_oi:
        grid_bpt[2].plot(xx_sf_oi, kewley_sf_oi(xx_sf_oi), 'r-', zorder=90)

    grid_bpt[0].plot(xx_comp_nii, kewley_comp_nii(xx_comp_nii), 'r-', zorder=90)

    grid_bpt[1].plot(xx_agn_sii, kewley_agn_sii(xx_agn_sii), 'b-', zorder=80)
    if use_oi:
        grid_bpt[2].plot(xx_agn_oi, kewley_agn_oi(xx_agn_oi), 'b-', zorder=80)

    # Adds captions
    grid_bpt[0].text(-1, -0.5, 'SF', ha='center', fontsize=12, zorder=100, color='c')
    grid_bpt[0].text(0.5, 0.5, 'AGN', ha='left', fontsize=12, zorder=100)
    grid_bpt[0].text(-0.08, -1.2, 'Comp', ha='left', fontsize=12, zorder=100, color='g')

    grid_bpt[1].text(-1.2, -0.5, 'SF', ha='center', fontsize=12, zorder=100)
    grid_bpt[1].text(-1, 1.2, 'Seyfert', ha='left', fontsize=12, zorder=100, color='r')
    grid_bpt[1].text(0.3, -1, 'LINER', ha='left', fontsize=12, zorder=100, color='m')

    if use_oi:
        grid_bpt[2].text(-2, -0.5, 'SF', ha='center', fontsize=12, zorder=100)
        grid_bpt[2].text(-1.5, 1, 'Seyfert', ha='left', fontsize=12, zorder=100)
        grid_bpt[2].text(-0.1, -1, 'LINER', ha='right', fontsize=12, zorder=100)

    # Sets the ticks, ticklabels, and other details
    xtick_limits = ((-2, 1), (-1.5, 1), (-2.5, 0.5))
    axes = [0, 1, 2] if use_oi else [0, 1]

    for ii in axes:

        grid_bpt[ii].get_xaxis().set_tick_params(direction='in')
        grid_bpt[ii].get_yaxis().set_tick_params(direction='in')

        grid_bpt[ii].set_xticks(np.arange(xtick_limits[ii][0], xtick_limits[ii][1] + 0.5, 0.5))
        grid_bpt[ii].set_xticks(np.arange(xtick_limits[ii][0],
                                          xtick_limits[ii][1] + 0.1, 0.1), minor=True)
        grid_bpt[ii].set_yticks(np.arange(-1.5, 2.0, 0.5))
        grid_bpt[ii].set_yticks(np.arange(-1.5, 1.6, 0.1), minor=True)

        grid_bpt[ii].grid(which='minor', alpha=0.2)
        grid_bpt[ii].grid(which='major', alpha=0.5)

        grid_bpt[ii].set_xlim(xtick_limits[ii][0], xtick_limits[ii][1])

        grid_bpt[ii].set_ylim(-1.5, 1.6)
        if use_oi:
            grid_bpt[ii].set_ylim(-1.5, 1.8)

        grid_bpt[ii].spines['top'].set_visible(True)

        if ii in [0, 1]:
            if not use_oi and ii == 1:
                continue
            grid_bpt[ii].get_xticklabels()[-1].set_visible(False)

    grid_bpt[0].set_ylabel(r'log([OIII]/H$\beta$)')

    grid_bpt[0].set_xlabel(r'log([NII]/H$\alpha$)')
    grid_bpt[1].set_xlabel(r'log([SII]/H$\alpha$)')
    if use_oi:
        grid_bpt[2].set_xlabel(r'log([OI]/H$\alpha$)')

    gal_bpt[0].grid(False)

    return fig, grid_bpt, gal_bpt[0]


def kewley_sf_nii(log_nii_ha):
    """Star forming classification line for log([NII]/Ha)."""
    return 0.61 / (log_nii_ha - 0.05) + 1.3


def kewley_sf_sii(log_sii_ha):
    """Star forming classification line for log([SII]/Ha)."""
    return 0.72 / (log_sii_ha - 0.32) + 1.3


def kewley_sf_oi(log_oi_ha):
    """Star forming classification line for log([OI]/Ha)."""
    return 0.73 / (log_oi_ha + 0.59) + 1.33


def kewley_comp_nii(log_nii_ha):
    """Composite classification line for log([NII]/Ha)."""
    return 0.61 / (log_nii_ha - 0.47) + 1.19


def kewley_agn_sii(log_sii_ha):
    """Seyfert/LINER classification line for log([SII]/Ha)."""
    return 1.89 * log_sii_ha + 0.76


def kewley_agn_oi(log_oi_ha):
    """Seyfert/LINER classification line for log([OI]/Ha)."""
    return 1.18 * log_oi_ha + 1.30


def bpt_kewley06(maps, snr_min=3, return_figure=True, use_oi=True, **kwargs):
    """Returns a classification of ionisation regions, as defined in Kewley+06.

    Makes use of the classification system defined by
    `Kewley et al. (2006) <https://ui.adsabs.harvard.edu/#abs/2006MNRAS.372..961K/abstract>`_
    to return classification masks for different ionisation mechanisms. If ``return_figure=True``,
    produces and returns a matplotlib figure with the classification plots (based on
    Kewley+06 Fig. 4) and the 2D spatial distribution of classified spaxels (i.e., a map of the
    galaxy in which each spaxel is colour-coded based on its emission mechanism).

    While it is possible to call this function directly, its normal use will be via the
    :func:`~marvin.tools.maps.Maps.get_bpt` method.

    Parameters:
        maps (a Marvin :class:`~marvin.tools.maps.Maps` object)
            The Marvin Maps object that contains the emission line maps to be used to determine
            the BPT classification.
        snr_min (float or dict):
            The signal-to-noise cutoff value for the emission lines used to generate the BPT
            diagram. If ``snr_min`` is a single value, that signal-to-noise will be used for all
            the lines. Alternatively, a dictionary of signal-to-noise values, with the
            emission line channels as keys, can be used.
            E.g., ``snr_min={'ha': 5, 'nii': 3, 'oi': 1}``. If some values are not provided,
            they will default to ``SNR>=3``. Note that the value ``sii`` will be applied to both
            ``[SII 6718]`` and ``[SII 6732]``.
        return_figure (bool):
            If ``True``, it also returns the matplotlib figure_ of the BPT diagram plot,
            which can be used to modify the style of the plot.
        use_oi (bool):
            If ``True``, uses the OI diagnostic diagram for spaxel classification.

    Returns:
        bpt_return:
            ``bpt_kewley06`` returns a dictionary of dictionaries of classification masks.
            The classification masks (not to be confused with bitmasks) are boolean arrays with the
            same shape as the Maps or Cube (without the spectral dimension) that can be used
            to select spaxels belonging to a certain excitation process (e.g., star forming).
            The returned dictionary has the following keys: ``'sf'`` (star forming), ``'comp'``
            (composite), ``'agn'``, ``'seyfert'``, ``'liner'``, ``'invalid'``
            (spaxels that are masked out at the DAP level), and ``'ambiguous'`` (good spaxels that
            do not fall in any  classification or fall in more than one). Each key provides access
            to a new dictionary with keys ``'nii'`` (for the constraints in the diagram NII/Halpha
            vs OIII/Hbeta),  ``'sii'`` (SII/Halpha vs OIII/Hbeta), ``'oi'`` (OI/Halpha vs
            OIII/Hbeta; only if ``use_oi=True``), and ``'global'``, which applies all the previous
            constraints at once. The ``'ambiguous'`` mask only contains the ``'global'``
            subclassification, while the ``'comp'`` dictionary only contains ``'nii'``.
            ``'nii'`` is not available for ``'seyfert'`` and ``'liner'``. All the global masks are
            unique (a spaxel can only belong to one of them) with the exception of ``'agn'``, which
            intersects with ``'seyfert'`` and ``'liner'``. Additionally, if ``return_figure=True``,
            ``bpt_kewley06`` will also return the matplotlib figure for the generated plot, and a
            list of axes for each one of the subplots.

    Example:
        >>> maps_8485_1901 = Maps(plateifu='8485-1901')
        >>> bpt_masks, fig, axes = bpt_kewley06(maps_8485_1901)

        Gets the global mask for star forming spaxels

        >>> sf = bpt_masks['sf']['global']

        Gets the seyfert mask based only on the SII/Halpha vs OIII/Hbeta diagnostics

        >>> seyfert_sii = bpt_masks['seyfert']['sii']

    """

    if 'snr' in kwargs:
        warnings.warn('snr is deprecated. Use snr_min instead. '
                      'snr will be removed in a future version of marvin',
                      MarvinDeprecationWarning)
        snr_min = kwargs.pop('snr')

    if len(kwargs.keys()) > 0:
        raise MarvinError('unknown keyword {0}'.format(list(kwargs.keys())[0]))

    # Gets the necessary emission line maps
    oiii = get_masked(maps, 'oiii_5008', snr=get_snr(snr_min, 'oiii'))
    nii = get_masked(maps, 'nii_6585', snr=get_snr(snr_min, 'nii'))
    ha = get_masked(maps, 'ha_6564', snr=get_snr(snr_min, 'ha'))
    hb = get_masked(maps, 'hb_4862', snr=get_snr(snr_min, 'hb'))
    oi = get_masked(maps, 'oi_6302', snr=get_snr(snr_min, 'oi'))

    sii_6718 = get_masked(maps, 'sii_6718', snr=get_snr(snr_min, 'sii'))
    sii_6732 = get_masked(maps, 'sii_6732', snr=get_snr(snr_min, 'sii'))
    sii = sii_6718 + sii_6732

    # Calculate masked logarithms
    log_oiii_hb = np.ma.log10(oiii / hb)
    log_nii_ha = np.ma.log10(nii / ha)
    log_sii_ha = np.ma.log10(sii / ha)
    log_oi_ha = np.ma.log10(oi / ha)

    # Calculates masks for each emission mechanism according to the paper boundaries.
    # The log_nii_ha < 0.05, log_sii_ha < 0.32, etc are necessary because the classification lines
    # diverge and we only want the region before the asymptota.
    sf_mask_nii = ((log_oiii_hb < kewley_sf_nii(log_nii_ha)) & (log_nii_ha < 0.05)).filled(False)
    sf_mask_sii = ((log_oiii_hb < kewley_sf_sii(log_sii_ha)) & (log_sii_ha < 0.32)).filled(False)
    sf_mask_oi = ((log_oiii_hb < kewley_sf_oi(log_oi_ha)) & (log_oi_ha < -0.59)).filled(False)
    sf_mask = sf_mask_nii & sf_mask_sii & sf_mask_oi if use_oi else sf_mask_nii & sf_mask_sii

    comp_mask = ((log_oiii_hb > kewley_sf_nii(log_nii_ha)) & (log_nii_ha < 0.05)).filled(False) & \
                ((log_oiii_hb < kewley_comp_nii(log_nii_ha)) & (log_nii_ha < 0.465)).filled(False)
    comp_mask &= (sf_mask_sii & sf_mask_oi) if use_oi else sf_mask_sii

    agn_mask_nii = ((log_oiii_hb > kewley_comp_nii(log_nii_ha)) |
                    (log_nii_ha > 0.465)).filled(False)
    agn_mask_sii = ((log_oiii_hb > kewley_sf_sii(log_sii_ha)) |
                    (log_sii_ha > 0.32)).filled(False)
    agn_mask_oi = ((log_oiii_hb > kewley_sf_oi(log_oi_ha)) |
                   (log_oi_ha > -0.59)).filled(False)
    agn_mask = agn_mask_nii & agn_mask_sii & agn_mask_oi if use_oi else agn_mask_nii & agn_mask_sii

    seyfert_mask_sii = agn_mask & (kewley_agn_sii(log_sii_ha) < log_oiii_hb).filled(False)
    seyfert_mask_oi = agn_mask & (kewley_agn_oi(log_oi_ha) < log_oiii_hb).filled(False)
    seyfert_mask = seyfert_mask_sii & seyfert_mask_oi if use_oi else seyfert_mask_sii

    liner_mask_sii = agn_mask & (kewley_agn_sii(log_sii_ha) > log_oiii_hb).filled(False)
    liner_mask_oi = agn_mask & (kewley_agn_oi(log_oi_ha) > log_oiii_hb).filled(False)
    liner_mask = liner_mask_sii & liner_mask_oi if use_oi else liner_mask_sii

    # The invalid mask is the combination of spaxels that are invalid in all of the emission maps
    invalid_mask_nii = ha.mask | oiii.mask | nii.mask | hb.mask
    invalid_mask_sii = ha.mask | oiii.mask | sii.mask | hb.mask
    invalid_mask_oi = ha.mask | oiii.mask | oi.mask | hb.mask

    invalid_mask = ha.mask | oiii.mask | nii.mask | hb.mask | sii.mask
    if use_oi:
        invalid_mask |= oi.mask

    # The ambiguous mask are spaxels that are not invalid but don't fall into any of the
    # emission mechanism classifications.
    ambiguous_mask = ~(sf_mask | comp_mask | seyfert_mask | liner_mask) & ~invalid_mask

    sf_classification = {'global': sf_mask,
                         'nii': sf_mask_nii,
                         'sii': sf_mask_sii}

    comp_classification = {'global': comp_mask,
                           'nii': comp_mask}

    agn_classification = {'global': agn_mask,
                          'nii': agn_mask_nii,
                          'sii': agn_mask_sii}

    seyfert_classification = {'global': seyfert_mask,
                              'sii': seyfert_mask_sii}

    liner_classification = {'global': liner_mask,
                            'sii': liner_mask_sii}

    invalid_classification = {'global': invalid_mask,
                              'nii': invalid_mask_nii,
                              'sii': invalid_mask_sii}

    ambiguous_classification = {'global': ambiguous_mask}

    if use_oi:
        sf_classification['oi'] = sf_mask_oi
        agn_classification['oi'] = agn_mask_oi
        seyfert_classification['oi'] = seyfert_mask_oi
        liner_classification['oi'] = liner_mask_oi
        invalid_classification['oi'] = invalid_mask_oi

    bpt_return_classification = {'sf': sf_classification,
                                 'comp': comp_classification,
                                 'agn': agn_classification,
                                 'seyfert': seyfert_classification,
                                 'liner': liner_classification,
                                 'invalid': invalid_classification,
                                 'ambiguous': ambiguous_classification}

    if not return_figure:
        return bpt_return_classification

    # Does all the plotting
    with plt.style.context('seaborn-darkgrid'):
        fig, grid_bpt, gal_bpt = _get_kewley06_axes(use_oi=use_oi)

    sf_kwargs = {'marker': 's', 's': 12, 'color': 'c', 'zorder': 50, 'alpha': 0.7, 'lw': 0.0,
                 'label': 'Star-forming'}
    sf_handler = grid_bpt[0].scatter(log_nii_ha[sf_mask], log_oiii_hb[sf_mask], **sf_kwargs)
    grid_bpt[1].scatter(log_sii_ha[sf_mask], log_oiii_hb[sf_mask], **sf_kwargs)

    comp_kwargs = {'marker': 's', 's': 12, 'color': 'g', 'zorder': 45, 'alpha': 0.7, 'lw': 0.0,
                   'label': 'Composite'}
    comp_handler = grid_bpt[0].scatter(log_nii_ha[comp_mask], log_oiii_hb[comp_mask],
                                       **comp_kwargs)
    grid_bpt[1].scatter(log_sii_ha[comp_mask], log_oiii_hb[comp_mask], **comp_kwargs)

    seyfert_kwargs = {'marker': 's', 's': 12, 'color': 'r', 'zorder': 40, 'alpha': 0.7, 'lw': 0.0,
                      'label': 'Seyfert'}
    seyfert_handler = grid_bpt[0].scatter(log_nii_ha[seyfert_mask], log_oiii_hb[seyfert_mask],
                                          **seyfert_kwargs)
    grid_bpt[1].scatter(log_sii_ha[seyfert_mask], log_oiii_hb[seyfert_mask], **seyfert_kwargs)

    liner_kwargs = {'marker': 's', 's': 12, 'color': 'm', 'zorder': 35, 'alpha': 0.7, 'lw': 0.0,
                    'label': 'LINER'}
    liner_handler = grid_bpt[0].scatter(log_nii_ha[liner_mask], log_oiii_hb[liner_mask],
                                        **liner_kwargs)
    grid_bpt[1].scatter(log_sii_ha[liner_mask], log_oiii_hb[liner_mask], **liner_kwargs)

    amb_kwargs = {'marker': 's', 's': 12, 'color': '0.6', 'zorder': 30, 'alpha': 0.7, 'lw': 0.0,
                  'label': 'Ambiguous '}
    amb_handler = grid_bpt[0].scatter(log_nii_ha[ambiguous_mask], log_oiii_hb[ambiguous_mask],
                                      **amb_kwargs)
    grid_bpt[1].scatter(log_sii_ha[ambiguous_mask], log_oiii_hb[ambiguous_mask], **amb_kwargs)

    if use_oi:
        grid_bpt[2].scatter(log_oi_ha[sf_mask], log_oiii_hb[sf_mask], **sf_kwargs)
        grid_bpt[2].scatter(log_oi_ha[comp_mask], log_oiii_hb[comp_mask], **comp_kwargs)
        grid_bpt[2].scatter(log_oi_ha[seyfert_mask], log_oiii_hb[seyfert_mask], **seyfert_kwargs)
        grid_bpt[2].scatter(log_oi_ha[liner_mask], log_oiii_hb[liner_mask], **liner_kwargs)
        grid_bpt[2].scatter(log_oi_ha[ambiguous_mask], log_oiii_hb[ambiguous_mask], **amb_kwargs)

    # Creates the legend
    grid_bpt[0].legend([sf_handler, comp_handler, seyfert_handler, liner_handler, amb_handler],
                       ['Star-forming', 'Composite', 'Seyfert', 'LINER', 'Ambiguous'], ncol=2,
                       loc='upper left', frameon=True, labelspacing=0.1, columnspacing=0.1,
                       handletextpad=0.1, fontsize=9)

    # Creates a RGB image of the galaxy, and sets the colours of the spaxels to match the
    # classification masks
    gal_rgb = np.zeros((ha.shape[0], ha.shape[1], 3), dtype=np.uint8)

    for ii in [1, 2]:  # Cyan
        gal_rgb[:, :, ii][sf_mask] = 255

    gal_rgb[:, :, 1][comp_mask] = 128  # Green

    gal_rgb[:, :, 0][seyfert_mask] = 255  # Red

    # Magenta
    gal_rgb[:, :, 0][liner_mask] = 255
    gal_rgb[:, :, 2][liner_mask] = 255

    for ii in [0, 1, 2]:
        gal_rgb[:, :, ii][invalid_mask] = 255  # White
        gal_rgb[:, :, ii][ambiguous_mask] = 169  # Grey

    # Shows the image.
    gal_bpt.imshow(gal_rgb, origin='lower', aspect='auto', interpolation='nearest')

    gal_bpt.set_xlim(0, ha.shape[1] - 1)
    gal_bpt.set_ylim(0, ha.shape[0] - 1)
    gal_bpt.set_xlabel('x [spaxels]')
    gal_bpt.set_ylabel('y [spaxels]')

    axes = grid_bpt.axes_all + [gal_bpt]

    # Adds custom method to create figure
    for ax in axes:
        setattr(ax.__class__, 'bind_to_figure', _bind_to_figure)

    return (bpt_return_classification, fig, axes)


def _bind_to_figure(self, fig=None):
    """Copies axes to a new figure.

    Uses ``marvin.utils.plot.utils.bind_to_figure`` with a number
    of tweaks.

    """

    new_figure = bind_to_figure(self, fig=fig)

    if new_figure.axes[0].get_ylabel() == '':
        new_figure.axes[0].set_ylabel('log([OIII]/H$\\beta$)')

    return new_figure



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-06-21 17:01:09
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last Modified time: 2018-10-17 00:22:19

from __future__ import absolute_import, division, print_function

import abc
import os
import time
import six

import marvin
import marvin.tools.plate
from marvin.core.exceptions import MarvinError
from marvin.utils.general import parseIdentifier
from astropy.io import fits
from functools import wraps

import sdss_access.path
import sdss_access.sync


__ALL__ = ['VACContainer', 'VACMixIn']


def check_for_vac(f):
    ''' Decorator to check for and download VAC '''
    @wraps(f)
    def decorated_function(inst, *args, **kwargs):
        if 'path' in kwargs and kwargs['path']:
            for kw in kwargs['path'].split('/'):
                if len(kw) == 0:
                    continue
                var, value = kw.split('=')
                kwargs[var] = value
        kwargs.pop('path')
        return f(inst, *args, **kwargs)
    return decorated_function


class VACContainer(object):

    def __repr__(self):
        return '<VACContainer ({0})>'.format(', '.join(map(repr, list(self))))

    def __dir__(self):
        props = []
        for value in self.__class__.__dict__.keys():
            if not value.startswith('_'):
                props.append(value)
        return props

    def __getitem__(self, value):
        return getattr(self, value)

    def __iter__(self):
        for value in self.__dir__():
            yield value


class VACMixIn(object, six.with_metaclass(abc.ABCMeta)):
    """MixIn  that allows VAC integration in Marvin.

    This parent class provides common tools for downloading data using
    sdss_access or directly from the sandbox. `~VACMixIn.get_vacs` returns a
    container with properties pointing to all the VACs that subclass from
    `.VACMixIn`. In general, VACs can be added to a class in the following way:

    .. code-block:: python

        from marvin.contrib.vacs.base import VACMixIn

        class Maps(MarvinToolsClass):

            def __init__(self, *args, **kwargs):

                ...

                self.vacs = VACMixIn.get_vacs(self)

    and then the VACs can be accessed as properties in ``my_map.vacs``.

    """
    # Set this is True on your VAC to exclude it from Marvin
    _hidden = False
    _hidden_for = None

    # The name and description of the VAC.
    name = None
    description = None
    url = None
    display_name = None
    
    # custom data container for VAC data in summary file(s)
    # used by tools.vacs.VACs
    data_container = None

    def __init__(self):

        if not sdss_access.sync.Access:
            raise MarvinError('sdss_access is not installed')
        else:
            self._release = marvin.config.release
            # is_public = 'DR' in self._release
            # rsync_release = self._release.lower() if is_public else None
            self.rsync_access = sdss_access.sync.Access(release=self._release)

        # file path for VAC summary file
        self.summary_file = None
        self.set_summary_file(marvin.config.release)

    def __repr__(self):
        return '<VAC (name={0}, description={1})>'.format(self.name, self.description)

    @abc.abstractmethod
    def get_target(self, parent_object):
        """Returns VAC data that matches the `parent_object` target.

        This method must be overridden in each subclass of `VACMixIn`. Details
        will depend on the exact implementation and the type of VAC, but in
        general each version of this method must:

        * Check whether the VAC file exists locally.
        * If it does not, download it using `~VACMixIn.download_vac`.
        * Open the file using the appropriate library.
        * Retrieve the VAC data matching ``parent_object``. Usually one will
          use attributes in ``parent_object`` such as ``.mangaid`` or
          ``.plateifu`` to perform the match.
        * Return the VAC data in whatever format is appropriate.

        """

        pass

    @staticmethod
    def get_vacs(parent_object):
        """Returns a container with all the VACs subclassing from `VACMixIn`.

        Because this method loops over ``VACMixIn.__subclasses__()``, all the
        class that inherit from `VACMixIn` and that must be included in the
        container need to have been imported before calling
        `~VACMixIn.get_vacs`.

        Parameters
        ----------
        parent_object : object
            The object to which the VACs are being attached. It will be passed
            to `~VACMixIn.get_target` when the subclass of `VACMixIn` is
            called.

        Returns
        -------
        vac_container : object
            An instance of a class that contains just a list of properties, one
            for to each on of the VACs that subclass from `VACMixIn`.

        """

        vac_container = VACContainer()

        for subvac in VACMixIn.__subclasses__():
            # Excludes VACs from showing up in Plate
            if issubclass(parent_object.__class__, marvin.tools.plate.Plate):
                continue

            # Only shows VACs if in the include list.
            if (hasattr(subvac, 'include') and subvac.include is not None and
                    not issubclass(parent_object.__class__, subvac.include)):
                continue

            # check if VAC is hidden
            if subvac._hidden and (not subvac._hidden_for or 
                                   (subvac._hidden_for == parent_object._release)):
                continue

            # We need to set sv=subvac in the lambda function to prevent
            # a cell-var-from-loop issue.
            if parent_object._release in subvac.version:
                setattr(VACContainer, subvac.name,
                        property(lambda self, sv=subvac: sv().get_target(parent_object)))

        return vac_container

    def download_vac(self, name=None, path_params={}, verbose=True):
        """Download the VAC using rsync and returns the local path."""

        if name is None:
            name = self.name

        assert name in self.rsync_access.templates, 'VAC path has not been set in the tree.'

        if verbose:
            marvin.log.info('downloading file for VAC {0!r}'.format(self.name))

        self.rsync_access.remote()
        self.rsync_access.add(name, **path_params)
        self.rsync_access.set_stream()
        self.rsync_access.commit()
        paths = self.rsync_access.get_paths()

        # adding a millisecond pause for download to finish and file existence to register
        time.sleep(0.001)

        return paths[0]  # doing this for single files, may need to change

    def get_path(self, name=None, path_params={}):
        """Returns the local VAC path or False if it does not exist."""

        if name is None:
            name = self.name

        # return the full local path to the file
        path = self.rsync_access.full(name, **path_params)
        return path
        # # check for and expand any wildcards present in the path_params
        # if self.rsync_access.any(name, **path_params):
        #     files = self.rsync_access.expand(name, **path_params)
        #     return files[0]
        # else:
        #     return False

    def file_exists(self, path=None, name=None, path_params={}):
        """Check whether a file exists locally"""

        # use the filepath if present
        if path:
            return os.path.exists(path)

        # otherwise use name and path_params
        if name is None:
            name = self.name

        if os.path.exists(self.get_path(name=name, path_params=path_params)):
            return True

        return False

    def check_vac(self, summary_file):
        ''' Checks the summary file for existence '''
        pass

    @abc.abstractmethod
    def set_summary_file(self, release):
        """ Sets the VAC summary file

        This method must be overridden in each subclass of `VACMixIn`. Details
        will depend on the exact implementation and the type of VAC, but in
        general each version of this method must:

        * Access the version of your VAC matching the current ``release``
        * Define a dictionary of keyword parameters that defines the `tree` path
        * Use `~VACMixIn.get_path` to construct the VAC path
        * Set that path to the `~VACMixIn.summary_file` attribute

        Setting a VAC summary file allows the `~marvin.tools.vacs.VACs` tool to load
        the full VAC data.  If the VAC does not contain a summary file, this method
        should `pass` or return `None`.

        """
        pass

    def update_path_params(self, params):
        ''' Update the path_params dictionary with additional parameters '''

        assert isinstance(params, dict), 'input parameters must be a dictionary'
        self.path_params.update(params)

    def get_ancillary_file(self, name, path_params={}):
        ''' Get a path to an ancillary VAC file '''

        path = self.get_path(name, path_params=path_params)
        if not path:
            path = self.download_vac(name, path_params=path_params)
        return path


class VACTarget(object):
    ''' Customization Class to allow for returning complex target data

    This parent class provides a framework for returning more complex data associated
    with a given target observation, for example ancillary spectral or image data.  In these
    cases, returning a target row from the main VAC summary file, or a simple dictionary of values
    may not be sufficient.  This class can be subclassed and customized to return any
    extra functionality or data.

    When used, this class provides convenient access to the underlying VAC data as well
    as a boolean to indicate if the given target is included in the VAC.

    Parameters:
        targetid (str):
            The target id, usually plateifu or mangaid.  Required.
        vacfile (str):
            The path to the VAC summary file.  Required.

    Attributes:
        targetid (str):
            The plateifu or mangaid target designation
        data (row):
            The extracted row VAC data for the provided targetid
        _data (HDU):
            the first data HDU of the summary VAC FITS file
        _indata (bool):
            A boolean indicating if the target is included in the VAC

    To use, subclass this class, add a new `__init__` method.  Make sure to call the original
    class's `__init__` method with `super`.

    .. code-block:: python

        from marvin.contrib.vacs.base import VACTarget

        class ExampleTarget(VACTarget):

            def __init__(self, targetid, vacfile):
                super(ExampleTarget, self).__init__(targetid, vacfile)

    Further customization can now be done, e.g. adding new parameters in the initializtion of the
    object, adding new methods or attributes, or overriding existing methods, e.g. to customize the
    return `data` attribute.

    To access a single HDU from the VAC, use the `_get_data()` method.  If you need to access the entire file,
    use the `_open_file()` method.

    '''

    def __init__(self, targetid, vacfile, **kwargs):
        self.targetid = targetid
        self._ttype = parseIdentifier(targetid)
        assert self._ttype in ['plateifu', 'mangaid'], 'Input targetid must be a valid plateifu or mangaid'
        self._vacfile = vacfile
        self._data = self._get_data(self._vacfile)
        self._indata = targetid in self._data[self._ttype]

    def __repr__(self):
        return 'Target({0})'.format(self.targetid)

    @property
    def data(self):
        ''' The data row from a VAC for a specific targetid '''
        if not self._indata:
            return "No data exists for {0}".format(self.targetid)

        idx = self._data[self._ttype] == self.targetid
        return self._data[idx]

    @staticmethod
    def _open_file(vacfile):
        ''' Opens the full FITS VAC file '''
        return fits.open(vacfile)

    def _get_data(self, vacfile=None, ext=1):
        ''' Get only the data from the VAC file from a given extension '''
        if not vacfile:
            vacfile = self._vacfile
        return fits.getdata(vacfile, ext)




# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-12-08 14:24:58
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-13 18:21:26

from __future__ import print_function, division, absolute_import
from flask_classful import FlaskView
from flask import request, current_app, session as current_session
from marvin.web.web_utils import parseSession, update_allowed, updateGlobalSession
from marvin.web.web_utils import check_access, check_request_for_release
import marvin
from brain.utils.general import build_routemap
from marvin.api.base import arg_validate as av


class BaseWebView(FlaskView):
    ''' This is the Base Web View for all pages '''

    def __init__(self, page):
        self.base = {}
        self.base['intro'] = 'Welcome to Marvin!'
        self.base['version'] = marvin.__version__
        self.update_title(page)
        self._endpoint = self._release = None
        self._drpver = self._dapver = None

    def before_request(self, *args, **kwargs):
        ''' this runs before every single request '''

        # check Flask request for release info but only when no session
        if 'release' not in current_session:
            check_request_for_release(request)

        # check login/access status and update global session
        check_access()
        updateGlobalSession()

        self.base['error'] = None
        self._logged_in = current_session.get('loginready', False)
        self._versions = update_allowed()
        self._endpoint = request.endpoint
        self._drpver, self._dapver, self._release = parseSession()

        # try to get a local version of the urlmap for the arg_validator
        if not av.urlmap:
            urlmap = build_routemap(current_app)
            marvin.config.urlmap = urlmap
            av.urlmap = urlmap

    def after_request(self, name, response):
        ''' this runs after every single request '''

        return response

    def update_title(self, page):
        ''' Update the title and page '''
        self.base['title'] = page.title().split('-')[0] if 'main' in page \
            else page.title().replace('-', ' | ')
        self.base['page'] = page

    def reset_dict(self, mydict, exclude=None):
        ''' resets the page dictionary '''
        mydict['error'] = self.base['error']
        exclude = exclude if isinstance(exclude, list) else [exclude]
        diffkeys = set(mydict) - set(self.base)
        for key, val in mydict.items():
            if key in diffkeys and (key not in exclude):
                mydict[key] = '' if isinstance(val, str) else None
        mydict['versions'] = self._versions
        mydict['release'] = self._release
        mydict['loggedin'] = self._logged_in



#!/usr/bin/env python
# encoding: utf-8

"""caching_query.py

Represent functions and classes
which allow the usage of Dogpile caching with SQLAlchemy.
Introduces a query option called FromCache.

The three new concepts introduced here are:

 * CachingQuery - a Query subclass that caches and
   retrieves results in/from dogpile.cache.
 * FromCache - a query option that establishes caching
   parameters on a Query
 * RelationshipCache - a variant of FromCache which is specific
   to a query invoked during a lazy load.
 * _params_from_query - extracts value parameters from
   a Query.

The rest of what's here are standard SQLAlchemy and
dogpile.cache constructs.

"""
from sqlalchemy.orm.interfaces import MapperOption
from sqlalchemy.orm.query import Query
from dogpile.cache.api import NO_VALUE


class CachingQuery(Query):
    """A Query subclass which optionally loads full results from a dogpile
    cache region.

    The CachingQuery optionally stores additional state that allows it to
    consult a dogpile.cache cache before accessing the database, in the form
    of a FromCache or RelationshipCache object.   Each of these objects
    refer to the name of a :class:`dogpile.cache.Region` that's been configured
    and stored in a lookup dictionary.  When such an object has associated
    itself with the CachingQuery, the corresponding :class:`dogpile.cache.Region`
    is used to locate a cached result.  If none is present, then the
    Query is invoked normally, the results being cached.

    The FromCache and RelationshipCache mapper options below represent
    the "public" method of configuring this state upon the CachingQuery.

    """

    def __init__(self, regions, *args, **kw):
        self.cache_regions = regions
        Query.__init__(self, *args, **kw)

    def __iter__(self):
        """override __iter__ to pull results from dogpile
           if particular attributes have been configured.

           Note that this approach does *not* detach the loaded objects from
           the current session. If the cache backend is an in-process cache
           (like "memory") and lives beyond the scope of the current session's
           transaction, those objects may be expired. The method here can be
           modified to first expunge() each loaded item from the current
           session before returning the list of items, so that the items
           in the cache are not the same ones in the current Session.

        """
        if hasattr(self, '_cache_region'):
            return self.get_value(createfunc=lambda: list(Query.__iter__(self)))
        else:
            return Query.__iter__(self)

    def _get_cache_plus_key(self):
        """Return a cache region plus key."""

        dogpile_region = self.cache_regions[self._cache_region.region]
        if self._cache_region.cache_key:
            key = self._cache_region.cache_key
        else:
            key = _key_from_query(self)
        return dogpile_region, key

    def invalidate(self):
        """Invalidate the cache value represented by this Query."""

        dogpile_region, cache_key = self._get_cache_plus_key()
        dogpile_region.delete(cache_key)

    def get_value(self, merge=True, createfunc=None,
                  expiration_time=None, ignore_expiration=False):
        """Return the value from the cache for this query.

        Raise KeyError if no value present and no
        createfunc specified.

        """
        dogpile_region, cache_key = self._get_cache_plus_key()

        # ignore_expiration means, if the value is in the cache
        # but is expired, return it anyway.   This doesn't make sense
        # with createfunc, which says, if the value is expired, generate
        # a new value.
        assert not ignore_expiration or not createfunc, \
            "Can't ignore expiration and also provide createfunc"

        if ignore_expiration or not createfunc:
            cached_value = dogpile_region.get(cache_key,
                                              expiration_time=expiration_time,
                                              ignore_expiration=ignore_expiration)
        else:
            cached_value = dogpile_region.get_or_create(
                                    cache_key,
                                    createfunc,
                                    expiration_time=expiration_time
                                )
        if cached_value is NO_VALUE:
            raise KeyError(cache_key)
        if merge:
            cached_value = self.merge_result(cached_value, load=False)
        return cached_value

    def set_value(self, value):
        """Set the value in the cache for this query."""

        dogpile_region, cache_key = self._get_cache_plus_key()
        dogpile_region.set(cache_key, value)

    def use_cache(self, backend='default'):
        ''' Adds the cache onto a Query instance
        
        Parameters:
            backend (str):
                Type of cache backend to use.  Can be 'null', 'default', or 'maps'.
        
        Returns:
            returns a SQLA query instance with caching turned on
        '''
        from marvin import marvindb
        from marvin.db.caching import regions
        assert backend in list(regions.keys()), 'backend must be a proper cache backend'

        return self.options(FromCache(backend)).options(*marvindb.cache_bits)


def query_callable(regions, query_cls=CachingQuery):
    def query(*arg, **kw):
        return query_cls(regions, *arg, **kw)
    return query


def _key_from_query(query, qualifier=None):
    """Given a Query, create a cache key.

    There are many approaches to this; here we use the simplest,
    which is to create an md5 hash of the text of the SQL statement,
    combined with stringified versions of all the bound parameters
    within it.     There's a bit of a performance hit with
    compiling out "query.statement" here; other approaches include
    setting up an explicit cache key with a particular Query,
    then combining that with the bound parameter values.

    """

    stmt = query.with_labels().statement
    compiled = stmt.compile()
    params = compiled.params

    # here we return the key as a long string.  our "key mangler"
    # set up with the region will boil it down to an md5.
    return " ".join(
                    [str(compiled)] +
                    [str(params[k]) for k in sorted(params)])


class FromCache(MapperOption):
    """Specifies that a Query should load results from a cache."""

    propagate_to_loaders = False

    def __init__(self, region="default", cache_key=None):
        """Construct a new FromCache.

        :param region: the cache region.  Should be a
        region configured in the dictionary of dogpile
        regions.

        :param cache_key: optional.  A string cache key
        that will serve as the key to the query.   Use this
        if your query has a huge amount of parameters (such
        as when using in_()) which correspond more simply to
        some other identifier.

        """
        self.region = region
        self.cache_key = cache_key

    def process_query(self, query):
        """Process a Query during normal loading operation."""
        query._cache_region = self


class RelationshipCache(MapperOption):
    """Specifies that a Query as called within a "lazy load"
       should load results from a cache."""

    propagate_to_loaders = True

    def __init__(self, attribute, region="default", cache_key=None):
        """Construct a new RelationshipCache.

        :param attribute: A Class.attribute which
        indicates a particular class relationship() whose
        lazy loader should be pulled from the cache.

        :param region: name of the cache region.

        :param cache_key: optional.  A string cache key
        that will serve as the key to the query, bypassing
        the usual means of forming a key from the Query itself.

        """
        self.region = region
        self.cache_key = cache_key
        self._relationship_options = {
            (attribute.property.parent.class_, attribute.property.key): self
        }

    def process_query_conditionally(self, query):
        """Process a Query that is used within a lazy loader.

        (the process_query_conditionally() method is a SQLAlchemy
        hook invoked only within lazyload.)

        """
        if query._current_path:
            mapper, prop = query._current_path[-2:]
            key = prop.key

            for cls in mapper.class_.__mro__:
                if (cls, key) in self._relationship_options:
                    relationship_option = self._relationship_options[(cls, key)]
                    query._cache_region = relationship_option
                    break

    def and_(self, option):
        """Chain another RelationshipCache option to this one.

        While many RelationshipCache objects can be specified on a single
        Query separately, chaining them together allows for a more efficient
        lookup during load.

        """
        self._relationship_options.update(option._relationship_options)
        return self



from flask import current_app, Blueprint, render_template, jsonify
from flask import session as current_session, request, redirect, url_for
from flask_classful import route
from marvin import config
from brain.api.base import processRequest
from marvin.utils.general.general import parseIdentifier
from marvin.api.base import arg_validate as av
from marvin.web.controllers import BaseWebView
from marvin.web.web_utils import setGlobalSession, set_session_versions, get_web_releases
from marvin.web.extensions import cache

from brain.utils.general import validate_user, get_db_user
from flask_login import current_user, login_user, logout_user

if config.db:
    from marvin import marvindb

index = Blueprint("index_page", __name__)


class Marvin(BaseWebView):
    route_base = '/'

    def __init__(self):
        super(Marvin, self).__init__('marvin-main')
        self.main = self.base.copy()

    def before_request(self, *args, **kwargs):
        super(Marvin, self).before_request(*args, **kwargs)
        self.reset_dict(self.main)

    def index(self):
        current_app.logger.info('Welcome to Marvin Web!')
        return render_template("index.html", **self.main)

    def quote(self):
        return 'getting quote'

    def status(self):
        return 'OK'

    @route('/test/')
    def test(self):
        return 'new test'

    @route('/versions/')
    def get_versions(self):
        vers = {'sess_vers': current_session['versions'], 'config_vers': list(get_web_releases().keys()),
                'access': config.access, 'release': config.release, 'session_release': current_session['release']}
        return jsonify(result=vers)

    @route('/session/')
    def get_session(self):
        extra = {'access': config.access}
        return jsonify(result=dict(current_session, **extra))

    @route('/clear/')
    def clear_session(self):
        # clear the session
        current_session.clear()
        # clear the cache
        cache.clear()
        return jsonify(result=dict(current_session))

    def database(self):
        onecube = marvindb.session.query(marvindb.datadb.Cube).order_by(marvindb.datadb.Cube.pk).first()
        return jsonify(result={'plate': onecube.plate, 'status': 1})

    @route('/galidselect/', methods=['GET', 'POST'], endpoint='galidselect')
    def galidselect(self):
        ''' Route that handle the Navbar plate/galaxy id search form '''
        args = av.manual_parse(self, request, use_params='index', required='galid')
        galid = args.get('galid', None)
        if not galid:
            # if not galid return main page
            return redirect(url_for('index_page.Marvin:index'))
        else:
            idtype = parseIdentifier(galid)
        # check the idtype
        if idtype == 'plateifu' or idtype == 'mangaid':
            return redirect(url_for('galaxy_page.Galaxy:get', galid=galid))
        elif idtype == 'plate':
            return redirect(url_for('plate_page.Plate:get', plateid=galid))
        else:
            return redirect(url_for('index_page.Marvin:index'))

    @route('/getgalidlist/', methods=['GET', 'POST'], endpoint='getgalidlist')
    def getgalidlist(self):
        ''' Retrieves the list of galaxy ids and plates for Bloodhound Typeahead '''

        if marvindb.datadb is None:
            out = ['', '', '']
            current_app.logger.info('ERROR: Problem with marvindb.datadb.  Cannot build galaxy id auto complete list.')
        else:
            cubes = (marvindb.session.query(marvindb.datadb.Cube.plate, marvindb.datadb.Cube.mangaid,
                                            marvindb.datadb.Cube.plateifu).join(marvindb.datadb.PipelineInfo,
                                                                                marvindb.datadb.PipelineVersion,
                                                                                marvindb.datadb.IFUDesign).
                     filter(marvindb.datadb.PipelineVersion.version == str(self._drpver)).all())
            out = [str(e) for l in cubes for e in l]

        out = list(set(out))
        out.sort()
        return jsonify(out)

    @route('/selectmpl/', methods=['GET', 'POST'], endpoint='selectmpl')
    def selectmpl(self):
        ''' Global selection of the MPL/DR versions '''
        args = av.manual_parse(self, request, use_params='index')
        version = args.get('release', None)
        set_session_versions(version)
        drpver, dapver = config.lookUpVersions(release=version)
        out = {'status': 1, 'msg': 'Success', 'current_release': version,
               'current_drpver': drpver, 'current_dapver': dapver}

        return jsonify(result=out)

    @route('/logout/', methods=['GET', 'POST'], endpoint='logout')
    def logout(self):
        ''' logout from the system
        '''

        result = {'logout': 'success'}

        if 'loginready' in current_session:
            ready = current_session.pop('loginready')

        if 'name' in current_session:
            name = current_session.pop('name')

        request.environ['REMOTE_USER'] = None
        config.access = 'public'
        set_session_versions(config.release)
        setGlobalSession()
        logout_user()

        return redirect(url_for('index_page.Marvin:index'))

    @route('/login/', methods=['GET', 'POST'], endpoint='login')
    def login(self):
        form = processRequest(request=request)
        result = {}
        username = form['username'].strip()
        password = form['password'].strip()

        # do nothing if already validated
        if current_user.is_authenticated:
            return jsonify(result=result)

        # validate the user with htpassfile or trac username
        is_valid, user, result = validate_user(username, password, request=request)

        # get User only if valid
        if is_valid:
            user = get_db_user(username, password, dbsession=marvindb.session, user_model=marvindb.datadb.User, request=request)
            if user and user.check_password(password):
                login_user(user)

        if is_valid:
            result['status'] = 1
            result['message'] = 'Login Successful!'
            current_session['name'] = user.username
            current_session['loginready'] = True
            config.access = 'collab'
            setGlobalSession()
        else:
            result['status'] = -1
            result['message'] = 'Login {0} is not valid!'.format(username)

        return jsonify(result=result)


Marvin.register(index)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-04-04 23:07:41
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-10-18 14:12:38

from __future__ import absolute_import, division, print_function

import copy

from marvin.utils.datamodel.maskbit import get_maskbits

from .base import DAPDataModel
from .MPL6 import GAU_MILESHC, HYB10, VOR10, MPL6_maps, MPL6_models, binid_properties


# MPL-7 DapDataModel goes here
MPL7 = DAPDataModel('2.2.1', aliases=['MPL-7', 'MPL7', 'DR15', 'DR16'],
                    bintypes=[HYB10, VOR10],
                    db_only=[HYB10],
                    templates=[GAU_MILESHC],
                    properties=MPL6_maps,
                    models=MPL6_models,
                    bitmasks=get_maskbits('MPL-7'),
                    default_bintype='HYB10',
                    default_template='GAU-MILESHC',
                    property_table='SpaxelProp7',
                    default_binid=copy.deepcopy(binid_properties[0]),
                    default_mapmask=['NOCOV', 'UNRELIABLE', 'DONOTUSE'],
                    qual_flag='DAPQUAL')



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-10-11 17:51:43
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-29 17:23:15

from __future__ import print_function, division, absolute_import

import marvin.tools
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from marvin import log

from .base import VACMixIn, VACTarget


class VMORPHOVAC(VACMixIn):
    """Provides access to the MaNGA-VISUAL-MORPHOLOGY VAC.

    VAC name: manga_visual_morpho

    URL: https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=manga-visual-morphologies-from-sdss-and-desi-images
    
    Description: A new morphology catalogue is presented in this VAC, based on a pure visual
    morphological classification. This catalogue contains the T-Type morphology, visual
    attributes (barred, edge-on, tidal debris) and the CAS parameters (Concentration, Asymmetry
    and Clumpiness; from the DESI images.

    Authors: J. Antonio Vazquez-Mata and Hector Hernandez-Toledo

    """
    #  hidden from DR17 until future notice
    _hidden = True
    _hidden_for = 'DR17'

    # Required parameters
    name = 'visual_morphology'
    description = 'Returns visual morphology data'
    version = {'DR16': '1.0.1', 'DR17': '2.0.1', 'MPL-11': '2.0.1'}
    display_name = 'Visual Morphology'
    url = 'https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=manga-visual-morphologies-from-sdss-and-desi-images'

    # optional Marvin Tools to attach your vac to
    include = (marvin.tools.cube.Cube, marvin.tools.maps.Maps)

    # Required method
    def set_summary_file(self, release):
        ''' Sets the path to the Visual Morphology summary file '''

        # define the variables to build a unique path to your VAC file
        self.path_params = {"vmver": self.version[release]}

        # get_path returns False if the files do not exist locally
        self.summary_file = self.get_path('mangaVmorpho', path_params=self.path_params)

    # Required method
    def get_target(self, parent_object):
        ''' Accesses VAC data for a specific target from a Marvin Tool object '''
        
        if parent_object.release == 'DR16':
            log.warning('You are accessing outdated DR16 data for this VAC.  This target has updated data in DR17. We recommend using the new data release instead.')

        # get any parameters you need from the parent object
        plateifu = parent_object.plateifu

        # download the vac from the SAS if it does not already exist locally
        if not self.file_exists(self.summary_file):
            self.summary_file = self.download_vac('mangaVmorpho', path_params=self.path_params)

        # get path to ancillary VAC files 
        if parent_object.release == 'DR16':
            # for DR16 SDSS/DESI mosaic images
            self.update_path_params({'plateifu': plateifu, 'survey': '*'})
            sdss_mos, desi_mos= self._get_mosaics(self.path_params)

            # create container for more complex return data
            vmdata = VizMorphTarget(plateifu, vacfile=self.summary_file, sdss=sdss_mos, desi=desi_mos)
        elif parent_object.release in ['DR17', 'MPL-11']:
            # for DR17 combined mosaic images
            self.update_path_params({'plateifu': plateifu})
            mos_mos = self._check_mosaic('mos', self.path_params)

            # create container for more complex return data
            vmdata = VizMorphTarget(plateifu, vacfile=self.summary_file, mos=mos_mos)
            
        return vmdata

    def _get_mosaics(self, path_params):
        ''' Get the mosaic images for SDSS and DESI surveys for DR16

        Parameters:
            path_params (dict):
                The sdss_access keyword parameters to define a file path

        Returns:
            The SDSS and DESI local image filepaths
        '''
        sdss_mosaic = self._check_mosaic('sdss', path_params)
        desi_mosaic = self._check_mosaic('desi', path_params)
        return sdss_mosaic, desi_mosaic

    def _check_mosaic(self, survey, path_params):
        ''' Get a mosaic image file for a survey path

        Checks for local existence of the mosaic image filepath.
        If it does not exists, it downloads it.

        Parameters:
            survey (str):
                The survey to download.  Either sdss or desi in DR16; or mos in DR17
            path_params (dict):
                The sdss_access keyword parameters to define a file path

        Returns:
            The mosaic image file path
        '''
        path_params['survey'] = survey
        mosaic = self.get_path('mangaVmorphoImgs', path_params=path_params)
        # download the mosaic file (downloads both surveys at once)
        if not self.file_exists(mosaic):
            pp = path_params.copy()
            pp['survey'] = '*'
            mosaics = self.download_vac('mangaVmorphoImgs', path_params=pp)
        # get the path again for the single survey
        mosaic = self.get_path('mangaVmorphoImgs', path_params=path_params)
        return mosaic


class VizMorphTarget(VACTarget):
    ''' A customized target class to also display morphology mosaics

    This class handles data from both the Visual Morphology summary file and the
    individual image files.  Row data from the summary file for the given target
    is returned via the `data` property.  Images can be displayed via
    the the `show_mosaic` method.

    Parameters:
        targetid (str):
            The plateifu or mangaid designation
        vacfile (str):
            The path of the VAC summary file
        sdss (str):
            The path to the DR16 SDSS image mosaic
        desi (str):
            The path to the DR16 DESI image mosaic
        mos (str):
            The path to the DR17 combined image mosaic
            
    Attributes:
        data:
            The target row data from the main VAC file
        targetid (str):
            The target identifier
    '''

    def __init__(self, targetid, vacfile, sdss=None, desi=None, mos=None):
        super(VizMorphTarget, self).__init__(targetid, vacfile)
        self._sdss_img = sdss
        self._desi_img = desi
        self._mos_img = mos

    def show_mosaic(self, survey=None):
        ''' Show the mosaic image for the given survey in DR16 or the combined in DR17

        Displays the mosaic image of visual morphology classification
        for the given survey as a Matplotlib Figure/Axis object.

        Parameters:
            survey (str):
                The survey name.  Can be either "sdss" or "desi" for DR16; or "mos" for DR17

        Returns:
            A matplotlib axis object
        '''
        
        #print('NOTE: For DR16, must specify either survey: sdss or desi. For DR17 must write: mos')

        if survey == 'sdss':
            impath = self._sdss_img
            fsize = (15,5)
        elif survey == 'desi':
            impath = self._desi_img
            fsize = (10,5)
        elif survey == 'mos':
            impath = self._mos_img
            fsize = (20,5)
        else:
            raise ValueError('survey must be either "sdss" or "desi" for DR16, or "mos" for DR17')

        imdata = mpimg.imread(impath)
        fig, ax = plt.subplots(figsize = fsize)
        ax.imshow(imdata)
        title = '{0} Mosaic'.format(survey.upper())
        fig.suptitle(title)
        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-10-31
# @Filename: base_quantity.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-08-12 13:51:26


from __future__ import absolute_import, division, print_function

import warnings

import astropy.units as units
import numpy

import marvin.core.exceptions
from marvin.tools.spaxel import Spaxel
from marvin.utils.general import maskbit
from marvin.utils.general.general import _sort_dir


class BinInfo(object):
    """Provides information about the bin associated with this quantity."""

    def __init__(self, spaxel=None, parent=None, datamodel=None):

        self._spaxel = spaxel
        self._parent = parent
        self._datamodel = datamodel

    def __repr__(self):

        return '<BinInfo (binid={0.binid}, n_spaxels={1})>'.format(
            self, numpy.sum(self.binid_mask))

    @property
    def binid_map(self):
        """Returns the binid Map associated to this quantity."""

        return self._parent.get_binid(self._datamodel)

    @property
    def binid(self):
        """Returns the binid associated to this quantity and spaxel."""

        return int(self.binid_map[self._spaxel.y, self._spaxel.x].value)

    @property
    def binid_mask(self):
        """Returns a mask of the spaxels with the same binid."""

        return self.binid_map.value == self.binid

    @property
    def is_binned(self):
        """Returns `True`` if the parent object is binned."""

        return self._parent.is_binned()

    def get_bin_spaxels(self, lazy=True):
        """Returns a list of the spaxels associated with this bin.

        Parameters
        ----------
        lazy : bool
            If ``True``, the spaxels returned will be lazy loaded. Spaxels
            can be fully loaded by calling their `~.Spaxel.load` method.

        Returns
        -------
        spaxels : list
            A list of all the `.Spaxel` instances associated with this
            quantity binid.

        """

        if self.binid < 0:
            raise marvin.core.exceptions.MarvinError(
                'coordinates ({}, {}) do not correspond to a valid binid.'.format(self._spaxel.x,
                                                                                  self._spaxel.y))

        spaxel_coords = zip(*numpy.where(self.binid_map.value == self.binid))
        spaxels = []

        for ii, jj in spaxel_coords:
            spaxels.append(Spaxel(x=jj, y=ii, plateifu=self._spaxel.plateifu,
                                  release=self._spaxel.release, cube=self._spaxel._cube,
                                  maps=self._spaxel._maps, modelcube=self._spaxel._modelcube,
                                  bintype=self._spaxel.bintype, template=self._spaxel.template,
                                  lazy=lazy))

        return spaxels


class QuantityMixIn(object):
    """A MixIn that provides common functionalities to Quantity classes."""

    def __dir__(self):

        return_list = _sort_dir(self, self.__class__)
        return_list += ['value']

        return return_list

    def _init_bin(self, spaxel=None, datamodel=None, parent=None):

        self.bin = BinInfo(spaxel=spaxel, datamodel=datamodel, parent=parent)

    @property
    def pixmask(self):
        """Maskbit instance for the pixmask flag.

        See :ref:`marvin-utils-maskbit` for documentation and
        `~marvin.utils.general.maskbit.Maskbit` for API reference.

        """

        assert self.pixmask_flag, 'pixmask flag not set'

        pixmask = maskbit.Maskbit(self.pixmask_flag)
        pixmask.mask = self.mask

        return pixmask

    @property
    def masked(self):
        """Return a masked array.

        If the `~QuantityMixIn.pixmask` is set, and the maskbit contains the
        ``DONOTUSE`` and ``NOCOV`` labels, the returned array will be masked
        for the values containing those bits. Otherwise, all values where the
        mask is greater than zero will be masked.

        """

        assert self.mask is not None, 'mask is not set'

        try:
            pixmask = self.pixmask
        except AssertionError:
            warnings.warn('pixmask not set. Applying full mask.',
                          marvin.core.exceptions.MarvinUserWarning)
            return numpy.ma.array(self.value, mask=(self.mask > 0))

        labels = pixmask.schema.label.tolist()
        if 'DONOTUSE' in labels and 'NOCOV' in labels:
            return numpy.ma.array(self.value,
                                  mask=self.pixmask.get_mask(['DONOTUSE', 'NOCOV']) > 0)
        elif 'DONOTUSE' in labels:
            return numpy.ma.array(self.value, mask=self.pixmask.get_mask('DONOTUSE') > 0)
        else:
            return numpy.ma.array(self.value, mask=(self.mask > 0))

    @property
    def error(self):
        """Compute the standard deviation of the measurement."""

        if hasattr(self, '_std') and self._std is not None:
            return self._std

        if self.ivar is None:
            return None

        numpy.seterr(divide='ignore')

        return numpy.sqrt(1. / self.ivar) * self.unit

    @property
    def snr(self):
        """Return the signal-to-noise ratio for each spaxel in the map."""

        return numpy.abs(self.value * numpy.sqrt(self.ivar))

    def descale(self):
        """Returns a copy of the object in which the scale is unity.

        Example:

            >>> dc.unit
            Unit("1e-17 erg / (Angstrom cm2 s spaxel)")
            >> dc[100, 15, 15]
            <DataCube 0.270078063011169 1e-17 erg / (Angstrom cm2 s spaxel)>
            >>> dc_descaled = dc.descale()
            >>> d_descaled.unit
            Unit("Angstrom cm2 s spaxel")
            >>> dc[100, 15, 15]
            <DataCube 2.70078063011169e-18 erg / (Angstrom cm2 s spaxel)>

        """

        if self.unit.scale == 1:
            return self

        value_descaled = self.value * self.unit.scale
        value_unit = units.CompositeUnit(1, self.unit.bases, self.unit.powers)

        if self.ivar is not None:
            ivar_descaled = self.ivar / (self.unit.scale ** 2)
        else:
            ivar_descaled = None

        return self.__class__(value_descaled, self.wavelength, unit=value_unit,
                              ivar=ivar_descaled, mask=self.mask)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: MPL10.py
# Project: dap
# Author: Brian Cherinka
# Created: Monday, 13th July 2020 2:49:16 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Monday, 13th July 2020 2:49:16 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
import copy
from astropy import units as u
from marvin.utils.datamodel.maskbit import get_maskbits
from .base import DAPDataModel, Template
from .base import Model, MultiChannelProperty

from .MPL8 import SPX, HYB10, VOR10, binid_properties, MPL8_models
from .MPL9 import MPL9_maps, MPL8_specindex_channels

# update Template for MPL-10
MILESHC_MASTARHC2 = Template('MILESHC-MASTARHC2',
                             description=('Stellar kinematic templates from the MILES library.  '
                                          'Stellar continuum template derived from a hierarchically '
                                          'clustered set of MaStar templates; used during emission-line fits.'))

# MPL-10 maps
updated_maps = copy.deepcopy(MPL9_maps)
remove = ['specindex_bcen', 'specindex_bcnt', 'specindex_rcen', 'specindex_rcnt', 'emline_tplsigma']
updated_maps = [m for m in updated_maps if m.name not in remove]

# new properties
MPL10_specindex_channels = copy.deepcopy(MPL8_specindex_channels)

new_properties = [
    MultiChannelProperty('specindex_bf', ivar=True, mask=True,
                         channels=MPL10_specindex_channels,
                         formats={'string': 'modified BF spectral index'},
                         description='Measurements of spectral indices using a modified definition '
                                     'from Burstein (1984) and Faber (1985).'),
    MultiChannelProperty('specindex_bf_corr', ivar=False, mask=False,
                         channels=MPL10_specindex_channels,
                         formats={'string': 'BF spectral index sigma correction',
                                  'latex': r'BF spectral index $\sigma$ correction'},
                         description='Stellar velocity dispersion corrections for the '
                         'BF spectral index measurements '),
    MultiChannelProperty('specindex_bf_model', ivar=False, mask=False,
                         channels=MPL10_specindex_channels,
                         formats={'string': 'Best-fit BF Index Measurement',
                                  'latex': r'Best-fit BF Index Measurement'},
                         description='BF index measurements made using the best-fitting '
                                     'model spectrum'),
    MultiChannelProperty('specindex_wgt', ivar=True, mask=True,
                         channels=MPL10_specindex_channels,
                         formats={'string': 'Spectral index weights'},
                         description='Index weights to use when calculating an aggregated '
                                     'index for many spaxels/bins.'),
    MultiChannelProperty('specindex_wgt_corr', ivar=False, mask=False,
                         channels=MPL10_specindex_channels,
                         formats={'string': 'Index weight sigma correction',
                                  'latex': r'Index weight $\sigma$ correction'},
                         description='Stellar velocity dispersion corrections for the index weights.'),
    MultiChannelProperty('specindex_wgt_model', ivar=False, mask=False,
                         channels=MPL10_specindex_channels,
                         formats={'string': 'Best-fit Index Weights',
                                  'latex': r'Best-fit Index Weights'},
                         description='The index weights measured on the best-fitting model spectrum.'),
]

MPL10_maps = updated_maps + new_properties

# MPL-10 models
lsf = Model('lsf', 'LSF', 'WAVE', extension_ivar=None,
            extension_mask=None,
            unit=u.Angstrom,
            scale=1, formats={'string': 'Dispersion pre-pixel'},
            description='Broadened pre-pixel dispersion solution (1sigma LSF)',
            binid=binid_properties[0])
MPL10_models = copy.deepcopy(MPL8_models)
MPL10_models.append(lsf)


# MPL-10 DapDataModel goes here
MPL10 = DAPDataModel('3.0.1', aliases=['MPL-10', 'MPL10'],
                     bintypes=[SPX, HYB10, VOR10],
                     db_only=[HYB10],
                     templates=[MILESHC_MASTARHC2],
                     properties=MPL10_maps,
                     models=MPL10_models,
                     bitmasks=get_maskbits('MPL-10'),
                     default_bintype='HYB10',
                     default_template='MILESHC-MASTARHC2',
                     property_table='SpaxelProp10',
                     default_binid=copy.deepcopy(binid_properties[0]),
                     default_mapmask=['NOCOV', 'UNRELIABLE', 'DONOTUSE'],
                     qual_flag='DAPQUAL')



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-07-28 17:26:41
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-19 22:56:31

from __future__ import absolute_import, division, print_function

import abc
import os
import re
import time
import warnings

import six

from marvin import config, log
from marvin.core.exceptions import MarvinError, MarvinMissingDependency, MarvinUserWarning
from marvin.utils.db import testDbConnection
from marvin.utils.general.general import mangaid2plateifu
try:
    from sdss_access.path import Path
except ImportError:
    Path = None

try:
    from sdss_access import Access
except ImportError:
    Access = None

__all__ = ['MMAMixIn']


class MMAMixIn(object, six.with_metaclass(abc.ABCMeta)):
    """A mixin that provides multi-modal data access.

    Add this mixin to any new class object to provide that class with
    the Multi-Modal Data Access System for using local files, database,
    or remote connection when initializing new objects.  See :ref:`decision tree <marvin-dma>`

    Parameters:
        input (str):
            A string that can be a filename, plate-ifu, or mangaid. It will be
            automatically identified based on its unique format. This argument
            is always the first one, so it can be defined without the keyword
            for convenience.
        filename (str):
            The path of the file containing the file to load. If set,
            ``input`` is ignored.
        mangaid (str):
            The mangaid of the file to load. If set, ``input`` is ignored.
        plateifu (str):
            The plate-ifu of the data cube to load. If set, ``input`` is
            ignored.
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See :ref:`mode-decision-tree`.
        data (:class:`~astropy.io.fits.HDUList`, SQLAlchemy object, or None):
            An astropy ``HDUList`` or a SQLAlchemy object, to be used for
            initialisation. If ``None``, the :ref:`normal <marvin-dma>`` mode
            will be used.
        release (str):
            The MPL/DR version of the data to use.
        drpall (str):
            The path to the
            `drpall <https://trac.sdss.org/wiki/MANGA/TRM/TRM_MPL-5/metadata#DRP:DRPall>`_
            file to use. If not set it will use the default path for the file
            based on the ``release``
        download (bool):
            If ``True``, the data will be downloaded on instantiation. See
            :ref:`marvin-download-objects`.
        ignore_db (bool):
            If ``True``, the local data-origin `db` will be ignored.

    Attributes:
        data (:class:`~astropy.io.fits.HDUList`, SQLAlchemy object, or dict):
            Depending on the access mode, ``data`` is populated with the
            |HDUList| from the FITS file, a
            `SQLAlchemy <http://www.sqlalchemy.org>`_ object, or a dictionary
            of values returned by an API call.
        data_origin ({'file', 'db', 'api'}):
            Indicates the origin of the data, either from a file, the DB, or
            an API call.
        filename (str):
            The path of the file used, if any.
        mangaid (str):
            The mangaid of the target.
        plateifu:
            The plateifu of the target
        release:
            The data release

    """

    def __init__(self, input=None, filename=None, mangaid=None, plateifu=None,
                 mode=None, data=None, release=None, drpall=None, download=None,
                 ignore_db=False):
        self.data = data
        self.data_origin = None
        self._ignore_db = ignore_db

        self.filename = filename
        self.mangaid = mangaid
        self.plateifu = plateifu

        self.mode = mode if mode is not None else config.mode

        self._release = release if release is not None else config.release

        self._drpver, self._dapver = config.lookUpVersions(release=self._release)
        self._drpall = config._getDrpAllPath(self._drpver) if drpall is None else drpall

        self._forcedownload = download if download is not None else config.download

        self._determine_inputs(input)

        assert self.mode in ['auto', 'local', 'remote']
        assert self.filename is not None or self.plateifu is not None, 'no inputs set.'

        self.datamodel = None
        self._set_datamodel()

        if self.mode == 'local':
            self._doLocal()
        elif self.mode == 'remote':
            self._doRemote()
        elif self.mode == 'auto':
            try:
                self._doLocal()
            except Exception as ee:

                if self.filename:
                    # If the input contains a filename we don't want to go into remote mode.
                    raise(ee)
                else:
                    log.debug('local mode failed. Trying remote now.')
                    self._doRemote()

        # Sanity check to make sure data_origin has been properly set.
        assert self.data_origin in ['file', 'db', 'api'], 'data_origin is not properly set.'

    def _determine_inputs(self, input):
        """Determines what inputs to use in the decision tree."""

        if input:

            assert self.filename is None and self.plateifu is None and self.mangaid is None, \
                'if input is set, filename, plateifu, and mangaid cannot be set.'

            assert isinstance(input, six.string_types), 'input must be a string.'

            input_dict = self._parse_input(input)

            if input_dict['plate'] is not None and input_dict['ifu'] is not None:
                self.plateifu = input
            elif input_dict['plate'] is not None and input_dict['ifu'] is None:
                self._plate = input
            elif input_dict['mangaid'] is not None:
                self.mangaid = input
            else:
                # Assumes the input must be a filename
                self.filename = input

        if self.filename is None and self.mangaid is None and self.plateifu is None:
            raise MarvinError('no inputs defined.')

        if self.filename:
            self.mangaid = None
            self.plateifu = None

            if self.mode == 'remote':
                raise MarvinError('filename not allowed in remote mode.')

            assert os.path.exists(self.filename), \
                'filename {} does not exist.'.format(str(self.filename))

        elif self.plateifu:
            assert not self.filename, 'invalid set of inputs.'

        elif self.mangaid:
            assert not self.filename, 'invalid set of inputs.'
            self.plateifu = mangaid2plateifu(self.mangaid,
                                             drpall=self._drpall,
                                             drpver=self._drpver)

        elif self._plate:
            assert not self.filename, 'invalid set of inputs.'

    @staticmethod
    def _parse_input(value):
        """Parses and input and determines plate, ifu, and mangaid."""

        # Number of IFUs per size
        n_ifus = {19: 2, 37: 4, 61: 4, 91: 2, 127: 5, 7: 12}

        return_dict = {'plate': None, 'ifu': None, 'mangaid': None}

        plateifu_pattern = re.compile(r'([0-9]{4,5})-([0-9]{4,9})')
        ifu_pattern = re.compile('(7|127|[0-9]{2})([0-9]{2})')
        mangaid_pattern = re.compile(r'[0-9]{1,3}-[0-9]+')
        plateid_pattern = re.compile('([0-9]{4,})(?!-)(?<!-)')

        plateid_match = re.match(plateid_pattern, value)
        plateifu_match = re.match(plateifu_pattern, value)
        mangaid_match = re.match(mangaid_pattern, value)

        # Check whether the input value matches the plateifu pattern
        if plateifu_match is not None:
            plate, ifu = plateifu_match.groups(0)

            # If the value matches a plateifu, checks that the ifu is a valid one.
            ifu_match = re.match(ifu_pattern, ifu)
            if ifu_match is not None:
                ifu_size, ifu_id = map(int, ifu_match.groups(0))
                if ifu_id <= n_ifus[ifu_size]:
                    return_dict['plate'] = plate
                    return_dict['ifu'] = ifu

        # Check whether this is a mangaid
        elif mangaid_match is not None:
            return_dict['mangaid'] = value

        # Check whether this is a plate
        elif plateid_match is not None:
            return_dict['plate'] = value

        return return_dict

    @staticmethod
    def _get_ifus(minis=None):
        ''' Returns a list of all the allowed IFU designs ids

        Parameters:
            minis (bool):
                If True, includes the mini-bundles

        Returns:
            A list of IFU designs

        '''

        # Number of IFUs per size
        n_ifus = {19: 2, 37: 4, 61: 4, 91: 2, 127: 5, 7: 12}

        # Pop the minis
        if not minis:
            __ = n_ifus.pop(7)

        ifus = ['{0}{1:02d}'.format(key, i + 1) for key, value in n_ifus.items() for i in range(value)]
        return ifus

    def _set_datamodel(self):
        """Sets the datamodel for this object. Must be overridden by each subclass."""
        pass

    def _doLocal(self):
        """Tests if it's possible to load the data locally."""

        if self.filename:

            if os.path.exists(self.filename):
                self.mode = 'local'
                self.data_origin = 'file'
            else:
                raise MarvinError('input file {0} not found'.format(self.filename))

        elif self.plateifu:

            from marvin import marvindb
            if marvindb:
                testDbConnection(marvindb.session)

            if marvindb and marvindb.db and not self._ignore_db:
                self.mode = 'local'
                self.data_origin = 'db'
            else:
                fullpath = self._getFullPath()

                if fullpath and os.path.exists(fullpath):
                    self.mode = 'local'
                    self.filename = fullpath
                    self.data_origin = 'file'
                else:
                    if self._forcedownload:
                        self.download()
                        self.data_origin = 'file'
                    else:
                        raise MarvinError('failed to retrieve data using '
                                          'input parameters.')

    def _doRemote(self):
        """Tests if remote connection is possible."""

        if self.filename:
            raise MarvinError('filename not allowed in remote mode.')
        else:
            self.mode = 'remote'
            self.data_origin = 'api'

    def download(self, pathType=None, **pathParams):
        """Download using sdss_access Rsync"""

        # # check for public release
        # is_public = 'DR' in self._release
        # rsync_release = self._release.lower() if is_public else None

        if not Access:
            raise MarvinError('sdss_access is not installed')
        else:
            access = Access(release=self._release)
            access.remote()
            access.add(pathType, **pathParams)
            access.set_stream()
            access.commit()
            paths = access.get_paths()
            # adding a millisecond pause for download to finish and file existence to register
            time.sleep(0.001)

            self.filename = paths[0]  # doing this for single files, may need to change

    @abc.abstractmethod
    def _getFullPath(self, pathType=None, url=None, **pathParams):
        """Returns the full path of the file in the tree.

        This method must be overridden by each subclass.

        """

        # # check for public release
        # is_public = 'DR' in self._release
        # ismpl = 'MPL' in self._release
        # path_release = self._release.lower() if is_public or ismpl else None

        if not Path:
            raise MarvinMissingDependency('sdss_access is not installed')
        else:
            path = Path(release=self._release)
            try:
                if url:
                    fullpath = path.url(pathType, **pathParams)
                else:
                    fullpath = path.full(pathType, **pathParams)
            except Exception as ee:
                warnings.warn('sdss_access was not able to retrieve the full path of the file. '
                              'Error message is: {0}'.format(str(ee)), MarvinUserWarning)
                fullpath = None

        return fullpath

    @property
    def release(self):
        """Returns the release."""

        return self._release

    @release.setter
    def release(self, value):
        """Fails when trying to set the release after instantiation."""

        raise MarvinError('the release cannot be changed once the object has been instantiated.')

    @property
    def plate(self):
        """Returns the plate id."""

        return int(self.plateifu.split('-')[0])

    @property
    def ifu(self):
        """Returns the IFU."""

        return int(self.plateifu.split('-')[1])



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: MPL9.py
# Project: dap
# Author: Brian Cherinka
# Created: Monday, 4th November 2019 2:32:54 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Monday, 11th November 2019 3:56:19 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import copy
from astropy import units as u
from marvin.utils.datamodel.maskbit import get_maskbits
from .base import DAPDataModel, Template, spaxel as spaxel_unit, reindex_channels
from .base import Model, Channel, MultiChannelProperty
from .MPL6 import SPX, HYB10, VOR10, MPL6_maps, binid_properties
from .MPL6 import MPL6_emline_channels, oii_channel, oiid_channel

from .MPL6 import MPL6_specindex_channels
from .MPL8 import SPX, HYB10, VOR10, MPL8_maps
from .MPL8 import MPL8_models, MPL8_emline_channels

# update Template for MPL-8
MILESHC_MASTARHC = Template('MILESHC-MASTARHC',
                            description=('Stellar kinematic templates from the MILES library.  '
                                         'Stellar continuum template derived from high-res spectra '
                                         'from the MaStar stellary library, used during emission-line fits.'))

emline_channels = copy.deepcopy(MPL8_emline_channels)
MPL8_specindex_channels = copy.deepcopy(MPL6_specindex_channels)
updated_maps = copy.deepcopy(MPL8_maps)

# new coo channels
coo_channel = Channel('r_h_kpc', formats={'string': 'R/(h/kpc)'}, idx=2)
for t in updated_maps:
    if t.name in ['spx_ellcoo', 'bin_lwellcoo']:
        t.append_channel(coo_channel, at_index=2, unit=u.kpc / u.h)


# new emission line channels
new_channels = [
    Channel('ariii_7137', formats={'string': 'ArIII 7137',
                                   'latex': r'$\forb{Ar\,III}\;\lambda 7137$'}, idx=28),
    Channel('ariii_7753', formats={'string': 'ArIII 7753',
                                   'latex': r'$\forb{Ar\,III}\;\lambda 7753$'}, idx=29),
    Channel('hei_7067', formats={'string': 'HeI 7067',
                                 'latex': r'$\forb{He\,I}\;\lambda 7067$'}, idx=27),
    Channel('h11_3771', formats={'string': 'H11 3771',
                                 'latex': r'$\forb{H\,11}\;\lambda 3771$'}, idx=3),
    Channel('h12_3751', formats={'string': 'H12 3751',
                                 'latex': r'$\forb{H\,12}\;\lambda 3751$'}, idx=2),
    Channel('siii_9071', formats={'string': 'SIII 9071',
                                 'latex': r'$\forb{S\,III}\;\lambda 9071$'}, idx=31),
    Channel('siii_9533', formats={'string': 'SIII 9533',
                                 'latex': r'$\forb{S\,III}\;\lambda 9533$'}, idx=33),
    Channel('peps_9548', formats={'string': 'P-epsilon 9548',
                                  'latex': r'P$\epsilon\;\lambda 9548$'}, idx=34),
    Channel('peta_9017', formats={'string': 'P-eta 9017',
                                  'latex': r'P$\eta\;\lambda 9017$'}, idx=30),
    Channel('pzet_9231', formats={'string': 'P-zeta 9231',
                                  'latex': r'P$\zeta\;\lambda 9231$'}, idx=32),
]

# create new MPL9 channels and reindex
tmp = emline_channels + new_channels
idx_list = ['oii_3729', 'h12_3751', 'h11_3771', 'hthe_3798', 'heta_3836', 'neiii_3869', 'hei_3889',
            'hzet_3890', 'neiii_3968', 'heps_3971', 'hdel_4102', 'hgam_4341', 'heii_4687', 'hb_4862',
            'oiii_4960', 'oiii_5008', 'ni_5199', 'ni_5201', 'hei_5877', 'oi_6302', 'oi_6365',
            'nii_6549', 'ha_6564', 'nii_6585', 'sii_6718', 'sii_6732', 'hei_7067', 'ariii_7137',
            'ariii_7753', 'peta_9017', 'siii_9071', 'pzet_9231', 'siii_9533', 'peps_9548']
MPL9_emline_channels = reindex_channels(tmp, names=idx_list, starting_idx=1)

# update existing emline properties with new channel list
for m in updated_maps:
    if 'emline' in m.name and '_fom' not in m.name:
        unit, scale = m[0].unit, m[0].unit.scale
        params = {'ivar': m[0].ivar, 'mask': m[0].mask, 'formats': m[0].formats,
                  'pixmask_flag': m[0].pixmask_flag}
        scale = unit.scale
        oii = oiid_channel if 'd' in m.channels[0].name else oii_channel
        chans = [oii] + MPL9_emline_channels
        m.update_channels(chans, unit=unit, scale=scale, **params)


# new properties
new_properties = [
    # New Emission Line extensions
    MultiChannelProperty('emline_sew_cnt', ivar=False, mask=False,
                         channels=[oiid_channel] + MPL9_emline_channels,
                         formats={'string': 'Continuum of Summed EW'},
                         unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Continuum used for summed-flux equivalent width measurement'),
    MultiChannelProperty('emline_gew_cnt', ivar=False, mask=False,
                         channels=[oii_channel] + MPL9_emline_channels,
                         formats={'string': 'Continuum of Gaussian-fit EW'},
                         unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Continuum used for Gaussian-fit equivalent width measurement'),
    # New Spectral Index extensions
    MultiChannelProperty('specindex_bcen', ivar=False, mask=False,
                         channels=MPL8_specindex_channels,
                         formats={'string': 'Flux-weighted center: blue sideband',
                                  'latex': r'Flux-weighted center: blue sideband'},
                         description='Flux-weighted center of the blue sideband in '
                                     'the spectral-index measurement'),
    MultiChannelProperty('specindex_bcnt', ivar=False, mask=False,
                         channels=MPL8_specindex_channels,
                         formats={'string': 'Continuum: blue sideband',
                                  'latex': r'Continuum: blue sideband'},
                         description='Continuum level in the blue sideband used in '
                                     'the spectral-index measurement'),
    MultiChannelProperty('specindex_rcen', ivar=False, mask=False,
                         channels=MPL8_specindex_channels,
                         formats={'string': 'Flux-weighted center: red sideband',
                                  'latex': r'Flux-weighted center: red sideband'},
                         description='Flux-weighted center of the red sideband in '
                                     'the spectral-index measurement'),
    MultiChannelProperty('specindex_rcnt', ivar=False, mask=False,
                         channels=MPL8_specindex_channels,
                         formats={'string': 'Continuum: red sideband',
                                  'latex': r'Continuum: red sideband'},
                         description='Continuum level in the red sideband used in '
                                     'the spectral-index measurement'),
    MultiChannelProperty('specindex_model', ivar=False, mask=False,
                         channels=MPL8_specindex_channels,
                         formats={'string': 'Best-fit Index Measurement',
                                  'latex': r'Best-fit Index Measurement'},
                         description='Index measurement made using the best-fitting '
                                     'stellar continuum model')
]

MPL9_maps = updated_maps + new_properties

# MPL-9 DapDataModel goes here
MPL9 = DAPDataModel('2.4.1', aliases=['MPL-9', 'MPL9'],
                    bintypes=[SPX, HYB10, VOR10],
                    db_only=[HYB10],
                    templates=[MILESHC_MASTARHC],
                    properties=MPL9_maps,
                    models=MPL8_models,
                    bitmasks=get_maskbits('MPL-9'),
                    default_bintype='HYB10',
                    default_template='MILESHC-MASTARHC',
                    property_table='SpaxelProp9',
                    default_binid=copy.deepcopy(binid_properties[0]),
                    default_mapmask=['NOCOV', 'UNRELIABLE', 'DONOTUSE'],
                    qual_flag='DAPQUAL')



#!/usr/bin/env python
# encoding: utf-8
#
'''
new mangadap db model classes - Sept 7 , for alternate schema

'''

from __future__ import division, print_function

import re

import marvin.db.models.DataModelClasses as datadb
import numpy as np
from astropy.io import fits
from marvin.core.caching_query import RelationshipCache
from marvin.db.database import db
from marvin.utils.datamodel.dap import datamodel
from sqlalchemy import Float, ForeignKeyConstraint, and_, case, cast, select
from sqlalchemy.engine import reflection
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import configure_mappers, relationship
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer


# ========================
# Define database classes
# ========================
Base = db.Base


def cameliseClassname(tableName):
    """Produce a camelised class name."""

    return str(tableName[0].upper() +
               re.sub(r'_([a-z])',
               lambda m: m.group(1).upper(), tableName[1:]))


def ClassFactory(name, tableName, BaseClass=db.Base, fks=None):
    tableArgs = [{'autoload': True, 'schema': 'mangadapdb'}]
    if fks:
        for fk in fks:
            tableArgs.insert(0, ForeignKeyConstraint([fk[0]], [fk[1]]))

    newclass = type(
        name, (BaseClass,),
        {'__tablename__': tableName,
         '__table_args__': tuple(tableArgs)})

    return newclass


class File(Base):
    __tablename__ = 'file'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<File (pk={0},name={1},tag={2})'.format(self.pk, self.filename, self.pipelineinfo.version.version)

    @property
    def is_map(self):
        return self.filetype.value == 'MAPS'

    @property
    def ftype(self):
        return self.filetype.value

    @property
    def partner(self):
        session = db.Session.object_session(self)
        return session.query(File).join(Structure, datadb.Cube, FileType).filter(
            Structure.pk == self.structure.pk, datadb.Cube.pk == self.cube.pk, FileType.pk != self.filetype.pk).one()

    @property
    def primary_header(self):
        primaryhdu = [h for h in self.hdus if h.extname.name == 'PRIMARY'][0]
        return primaryhdu.header

    @property
    def flux_header(self):
        ftype = self.filetype.value
        name = 'FLUX' if ftype == 'LOGCUBE' else 'EMLINE_GFLUX'
        fluxhdu = [h for h in self.hdus if h.extname.name == name][0]
        return fluxhdu.header

    @hybrid_property
    def quality(self):
        hdr = self.primary_header
        bits = hdr.get('DAPQUAL', None)
        if bits:
            return int(bits)
        else:
            return None

    @quality.expression
    def quality(cls):
        return select([HeaderValue.value.cast(Integer)]).\
                      where(and_(HeaderKeyword.pk==HeaderValue.header_keyword_pk,
                                 HduToHeaderValue.header_value_pk==HeaderValue.pk,
                                 HduToHeaderValue.hdu_pk==Hdu.pk,
                                 Hdu.file_pk==cls.pk,
                                 HeaderKeyword.name.ilike('DAPQUAL')
                        )).\
                      label('dapqual')


class CurrentDefault(Base):
    __tablename__ = 'current_default'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<CurrentDefault (pk={0},name={1})'.format(self.pk, self.filename)


class FileType(Base):
    __tablename__ = 'filetype'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<FileType (pk={0},value={1})'.format(self.pk, self.value)


class Hdu(Base):
    __tablename__ = 'hdu'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<Hdu (pk={0}, name={1}, ext={2})'.format(self.pk, self.name, self.extno)

    @property
    def header(self):
        '''Returns an astropy header'''

        session = db.Session.object_session(self)
        data = session.query(HeaderKeyword.name, HeaderValue.value,
                             HeaderValue.comment).join(HeaderValue, HduToHeaderValue).filter(
            HduToHeaderValue.header_value_pk == HeaderValue.pk,
            HduToHeaderValue.hdu_pk == self.pk).all()

        hdr = fits.Header(data)
        return hdr

    def header_to_dict(self):
        '''Returns a simple python dictionary header'''

        vals = self.header_values
        hdrdict = {str(val.keyword.name): val.value for val in vals}
        return hdrdict

    @property
    def name(self):
        return self.extname.name


class ExtName(Base):
    __tablename__ = 'extname'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<ExtName (pk={0}, name={1})'.format(self.pk, self.name)


class ExtType(Base):
    __tablename__ = 'exttype'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<ExtType (pk={0}, name={1})'.format(self.pk, self.name)


class ExtCol(Base):
    __tablename__ = 'extcol'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<ExtCol (pk={0}, name={1})'.format(self.pk, self.name)


class HduToExtCol(Base):
    __tablename__ = 'hdu_to_extcol'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<HduToExtcol (pk={0})'.format(self.pk)


class HduToHeaderValue(Base):
    __tablename__ = 'hdu_to_header_value'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<HduToHeaderValue (pk={0})'.format(self.pk)


class HeaderValue(Base):
    __tablename__ = 'header_value'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<HeaderValue (pk={0})'.format(self.pk)


class HeaderKeyword(Base):
    __tablename__ = 'header_keyword'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<HeaderKeyword (pk={0})'.format(self.pk)


class Structure(Base):
    __tablename__ = 'structure'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<Structure (pk={0}, bintype={1}, template={3})'.format(self.pk, self.bintype.name, self.template_kin.name)


class BinId(Base):
    __tablename__ = 'binid'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<BinId (pk={0}, id={1})'.format(self.pk, self.id)


def HybridRatio(line1, line2):
    ''' produces emission line ratio hybrid properties '''

    @hybrid_property
    def hybridRatio(self):

        if type(line1) == tuple:
            myline1 = getattr(self, line1[0]) + getattr(self, line1[1])
        else:
            myline1 = getattr(self, line1)

        if getattr(self, line2) > 0:
            return myline1 / getattr(self, line2)
        else:
            return -999.

    @hybridRatio.expression
    def hybridRatio(cls):

        if type(line1) == tuple:
            myline1 = getattr(cls, line1[0]) + getattr(cls, line1[1])
        else:
            myline1 = getattr(cls, line1)

        return cast(case([(getattr(cls, line2) > 0., myline1 / getattr(cls, line2)),
                          (getattr(cls, line2) == 0., -999.)]), Float)

    return hybridRatio


class SpaxelAtts(object):
    ''' New class to add attributes to all SpaxelProp classes '''
    pass

setattr(SpaxelAtts, 'ha_to_hb', HybridRatio('emline_gflux_ha_6564', 'emline_gflux_hb_4862'))
setattr(SpaxelAtts, 'nii_to_ha', HybridRatio('emline_gflux_nii_6585', 'emline_gflux_ha_6564'))
setattr(SpaxelAtts, 'oiii_to_hb', HybridRatio('emline_gflux_oiii_5008', 'emline_gflux_hb_4862'))
setattr(SpaxelAtts, 'oi_to_ha', HybridRatio('emline_gflux_oi_6302', 'emline_gflux_ha_6564'))
setattr(SpaxelAtts, 'sii_to_ha', HybridRatio(('emline_gflux_sii_6718', 'emline_gflux_sii_6732'), 'emline_gflux_ha_6564'))


def spaxel_factory(classname, clean=None):
    ''' class factory for the spaxelprop tables '''

    if clean:
        classname = 'Clean{0}'.format(classname)
    tablename = classname.lower()

    params = {'__tablename__': tablename, '__table_args__': {'autoload': True, 'schema': 'mangadapdb'}}
    if clean:
        params.update({'pk': Column(Integer, primary_key=True)})

    def newrepr(self):
        return '<{2} (pk={0}, file={1})'.format(self.pk, self.file_pk, classname)

    try:
        newclass = type(classname, (Base, SpaxelAtts,), params)
        newclass.__repr__ = newrepr
    except Exception as e:
        newclass = None

    return newclass


# create the (clean)spaxel models from the DAP datamodel
for dm in datamodel:
    classname = datamodel[dm].property_table
    newclass = spaxel_factory(classname)
    cleanclass = spaxel_factory(classname, clean=True)
    if newclass:
        locals()[classname] = newclass
    if cleanclass:
        locals()[cleanclass.__name__] = cleanclass


class BinMode(Base):
    __tablename__ = 'binmode'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<BinMode (pk={0}, name={1})'.format(self.pk, self.name)


class BinType(Base):
    __tablename__ = 'bintype'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<BinType (pk={0}, name={1})'.format(self.pk, self.name)


class ExecutionPlan(Base):
    __tablename__ = 'executionplan'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<ExecutionPlan (pk={0}, id={1})'.format(self.pk, self.id)


class Template(Base):
    __tablename__ = 'template'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<Template (pk={0}, name={1}, id={2})'.format(self.pk, self.name, self.id)


class ModelCube(Base):
    __tablename__ = 'modelcube'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<ModelCube (pk={0}, file={1})'.format(self.pk, self.file_pk)

    def get3DCube(self, extension='flux'):
        """Returns a 3D array of ``extension`` from the modelcube spaxels.

        For example, ``modelcube.get3DCube('flux')`` will return the original
        flux cube with the same ordering as the FITS data cube.

        Note that this method seems to be really slow retrieving arrays (this
        is especially serious for large IFUs).

        """

        session = db.Session.object_session(self)
        spaxels = session.query(getattr(ModelSpaxel, extension)).filter(
            ModelSpaxel.modelcube_pk == self.pk).order_by(ModelSpaxel.x, ModelSpaxel.y).all()

        # Assumes cubes are always square (!)
        nx = ny = int(np.sqrt(len(spaxels)))
        nwave = len(spaxels[0][0])

        spArray = np.array(spaxels)

        return spArray.transpose().reshape((nwave, ny, nx)).transpose(0, 2, 1)


class ModelSpaxel(Base):
    __tablename__ = 'modelspaxel'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<ModelSpaxel (pk={0}, mc={1})'.format(self.pk, self.modelcube_pk)


class RedCorr(Base):
    __tablename__ = 'redcorr'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<RedCorr (pk={0}, modelcube={1})'.format(self.pk, self.modelcube_pk)


class DapAll(Base):
    __tablename__ = 'dapall'
    __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

    def __repr__(self):
        return '<DapAll (pk={0}, file={1})'.format(self.pk, self.file_pk)


if 'testtable' in db.engine.table_names('mangadapdb'):
    class TestTable(Base):
        __tablename__ = 'testtable'
        __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

        file = relationship(File, backref="testtable")

        def __repr__(self):
            return ('<TestTable pk={0}, file={1}>'.format(self.pk, self.file_pk))

if 'cleanspaxeltest' in db.engine.table_names('mangadapdb'):
    class CleanSpaxelTest(Base):
        __tablename__ = 'cleanspaxeltest'
        __table_args__ = {'autoload': True, 'schema': 'mangadapdb'}

        pk = Column(Integer, primary_key=True)
        file = relationship(File, backref="cleanspaxeltest")

        def __repr__(self):
            return ('<CleanSpaxelTest pk={0}, file={1}, val={2}>'.format(self.pk, self.file_pk, self.value))

# -----
# Buld Relationships
# -----

File.cube = relationship(datadb.Cube, backref='dapfiles')
File.pipelineinfo = relationship(datadb.PipelineInfo, backref='dapfiles')
File.filetype = relationship(FileType, backref='files')
File.structure = relationship(Structure, backref='files')
CurrentDefault.file = relationship(File, uselist=False, backref='current_default')
Hdu.file = relationship(File, backref='hdus')
Hdu.extname = relationship(ExtName, backref='hdus')
Hdu.exttype = relationship(ExtType, backref='hdus')
Hdu.extcols = relationship(ExtCol, secondary=HduToExtCol.__table__, backref='hdus')
Hdu.header_values = relationship(HeaderValue, secondary=HduToHeaderValue.__table__, backref='hdus')
HeaderValue.keyword = relationship(HeaderKeyword, backref='values')
DapAll.file = relationship(File, uselist=False, backref='dapall')

Structure.executionplan = relationship(ExecutionPlan, backref='structures')
Structure.binmode = relationship(BinMode, backref='structures')
Structure.bintype = relationship(BinType, backref='structures')
Structure.template_kin = relationship(Template, foreign_keys=[Structure.template_kin_pk], backref='structures_kin')
Structure.template_pop = relationship(Template, foreign_keys=[Structure.template_pop_pk], backref='structures_pop')

insp = reflection.Inspector.from_engine(db.engine)

# add foreign key relationships on (Clean)SpaxelProp classses to File
spaxel_tables = {k: v for k, v in locals().items() if 'SpaxelProp' in k or 'CleanSpaxelProp' in k}

for classname, class_model in spaxel_tables.items():
    fks = insp.get_foreign_keys(class_model.__table__.name, schema='mangadapdb')
    if fks:
        backname = classname.lower().replace('prop', 'props')
        class_model.file = relationship(File, backref=backname)

fks = insp.get_foreign_keys(ModelCube.__table__.name, schema='mangadapdb')
if fks:
    ModelCube.file = relationship(File, backref='modelcube', uselist=False)

fks = insp.get_foreign_keys(ModelSpaxel.__table__.name, schema='mangadapdb')
if fks:
    ModelSpaxel.modelcube = relationship(ModelCube, backref='modelspaxels')

fks = insp.get_foreign_keys(RedCorr.__table__.name, schema='mangadapdb')
if fks:
    RedCorr.modelcube = relationship(ModelCube, backref='redcorr')

# ---------------------------------------------------------
# Test that all relationships/mappings are self-consistent.
# ---------------------------------------------------------

try:
    configure_mappers()
except RuntimeError as error:
    print('DapModelClasses, Error during configure_mapper: {0}'.format(error))

dap_cache = RelationshipCache(File.cube).\
               and_(RelationshipCache(File.pipelineinfo)).\
               and_(RelationshipCache(File.filetype)).\
               and_(RelationshipCache(File.structure)).\
               and_(RelationshipCache(ModelCube.file)).\
               and_(RelationshipCache(ModelSpaxel.modelcube)).\
               and_(RelationshipCache(RedCorr.modelcube))
# add the SpaxelProp models to the cache
for classname, class_model in spaxel_tables.items():
    dap_cache = dap_cache.and_(RelationshipCache(class_model.file))

# delete extra classes from the various loops
del class_model
del cleanclass
del newclass



#!/usr/bin/env python
# encoding: utf-8


# Created by Brian Cherinka on 2016-05-18 15:08:29
# Licensed under a 3-clause BSD license.

# Revision History:
#     Initial Version: 2016-05-18 15:08:29 by Brian Cherinka
#     Last Modified On: 2016-05-18 15:08:29 by Brian


from __future__ import print_function
from __future__ import division
from flask import jsonify
from flask_classful import route
from marvin.tools.plate import Plate
from marvin.api.base import BaseView, arg_validate as av


def _getPlate(plateid, nocubes=None, **kwargs):
    ''' Get a Plate Marvin Object '''
    plate = None
    results = {}

    # Pop the release to remove a duplicate input to Maps
    release = kwargs.pop('release', None)

    if not str(plateid).isdigit():
        results['error'] = 'Error: plateid is not a numeric value'
        return plate, results

    try:
        plate = Plate(plate=plateid, nocubes=nocubes, mode='local', release=release)
    except Exception as e:
        results['error'] = 'Failed to retrieve Plate for id {0}: {1}'.format(plateid, str(e))
    else:
        results['status'] = 1

    return plate, results


class PlateView(BaseView):
    """Class describing API calls related to plates."""

    route_base = '/plates/'

    def index(self):
        '''Returns general maps info

        .. :quickref: Plate; Get general plate info

        :form release: the release of MaNGA data
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: data message
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/plates/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": "this is a plate!"
           }

        '''
        self.results['status'] = 1
        self.results['data'] = 'this is a plate!'
        return jsonify(self.results)

    @route('/<plateid>/', methods=['GET', 'POST'], endpoint='getPlate')
    @av.check_args()
    def get(self, args, plateid):
        """Retrieve basic info about a plate

        .. :quickref: Plate; Get a plate given a plateid

        :param plateid: The plateid you want
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string plateid: the plateid
        :json dict header: the cube header as a dict
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/plate/8485/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"plateid": "8485",
                    "header": {"AIRMSMAX": "1.07643", "AIRMSMED": "1.04336", "AIRMSMIN": "1.03694", ... }
              }
           }

        """
        args = self._pop_args(args, arglist=['plateid'])
        plate, results = _getPlate(plateid, nocubes=True, **args)
        self.update_results(results)

        if not isinstance(plate, type(None)):
            # For now we don't return anything here, maybe later.

            platedict = {'plateid': plateid, 'header': plate._hdr}
            self.results['data'] = platedict

        return jsonify(self.results)

    @route('/<plateid>/cubes/', methods=['GET', 'POST'], endpoint='getPlateCubes')
    @av.check_args()
    def getPlateCubes(self, args, plateid):
        """Returns a list of all the cubes for this plate

        .. :quickref: Plate; Get a list of all cubes for this plate

        :param plateid: The plateid you want
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json list plateifus: a list of plate-ifus for this plate
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/plate/8485/cubes/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"plateifus": ["8485-12701","8485-12702","8485-1901","8485-1902", ...]
              }
           }

        """

        args = self._pop_args(args, arglist=['plateid'])
        plate, results = _getPlate(plateid, **args)
        self.update_results(results)

        if not isinstance(plate, type(None)):
            plateifus = [cube.plateifu for cube in plate]
            self.results['data'] = {'plateifus': plateifus}

        return jsonify(self.results)



#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-04-08 14:31:34
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-04-08 14:31:34 by Brian Cherinka
    Last Modified On: 2016-04-08 14:31:34 by Brian

'''
from __future__ import division, print_function

import os
import functools
import numpy as np
from itertools import groupby
from operator import itemgetter
from brain.utils.general.general import convertIvarToErr
from flask import Blueprint, jsonify, render_template, request
from flask import session as current_session, g
from flask_classful import route
from astropy.wcs import WCS

import marvin
from marvin.api.base import arg_validate as av
from marvin.core import marvin_pickle
from marvin.core.caching_query import FromCache
from marvin.core.exceptions import MarvinError
from marvin.tools.cube import Cube
from marvin.utils.datamodel.dap import datamodel
from marvin.utils.general.general import (_db_row_to_dict, convertImgCoords, getDapRedux,
                                          getDefaultMapPath, parseIdentifier, target_status,
                                          get_manga_image, convertCoords)
from marvin.utils.general.maskbit import Maskbit
from marvin.web.controllers import BaseWebView
from marvin.web.extensions import cache


try:
    from sdss_access.path import Path
except ImportError as e:
    Path = None

galaxy = Blueprint("galaxy_page", __name__)


def get_flagged_regions(data, value=None):
    ''' Retrieves bad pixel regions in a spectrum

    Searches an input mask for a given value, looks for
    regions of consecutive bad values and returns the
    min and max index bound for each consectuive region.

    Parameters:
        data (list):
            An array of mask values
        value (int):
            The maskbit value to search on

    Returns:
        A list of bad region bounds

    Example:
        >>> data = [0, 1, 1, 1, 0, 0, 2, 1, 1]
        >>> bad = get_flagged_region(data, value=1)
        >>> print(bad)
        >>> [[1, 3], [7,8]]
    '''
    chunks = []
    data = np.where(data == value)[0]
    data_list = data.tolist()
    if not data_list:
        return chunks
    for k, g in groupby(enumerate(data_list), lambda ix: ix[0] - ix[1]):
        tmp = (list(map(itemgetter(1), g)))
        chunks.append([min(tmp), max(tmp)])
    return chunks


def getWebSpectrum(cube, x, y, xyorig=None, byradec=False):
    ''' Get and format a spectrum for the web '''
    webspec = None
    badspots = []
    default_bintype = datamodel[cube.release].default_bintype.name
    has_models = cube.data.has_modelspaxels(name=default_bintype) if hasattr(cube.data, 'has_modelspaxels') else False

    # set the spaxel kwargs
    kwargs = {'xyorig': xyorig, 'maps': False}
    kwargs['modelcube'] = has_models
    if byradec:
        kwargs.update({'ra': x, 'dec': y})
    else:
        kwargs.update({'x': x, 'y': y})

    # get the spaxel
    try:
        spaxel = cube.getSpaxel(**kwargs)
    except Exception as e:
        specmsg = 'Could not get spaxel: {0}'.format(e)
    else:
        # get error and wavelength
        error = convertIvarToErr(spaxel.flux.ivar)
        wave = spaxel.flux.wavelength

        # retrieve the index bounds for DONOTUSE pixel regions
        badspots = get_flagged_regions(spaxel.flux.pixmask.get_mask(
            'DONOTUSE'), value=spaxel.flux.pixmask.labels_to_value('DONOTUSE'))

        # try to get the model flux
        try:
            modelfit = spaxel.full_fit
        except Exception as e:
            modelfit = None

        # make input array for Dygraph
        if not isinstance(modelfit, type(None)):
            webspec = [[wave.value[i], [s, error[i]], [modelfit.value[i], 0.0]] for i, s in enumerate(spaxel.flux.value)]
        else:
            webspec = [[wave.value[i], [s, error[i]]] for i, s in enumerate(spaxel.flux.value)]

        specmsg = "Spectrum in Spaxel (j, i)=({2},{3}) at RA, Dec = ({0}, {1})".format(spaxel.ra, spaxel.dec, spaxel.x, spaxel.y)

    return webspec, specmsg, badspots


def getWebMap(cube, parameter='emline_gflux', channel='ha_6564',
              bintype=None, template=None):
    ''' Get and format a map for the web '''
    if channel:
        name = '{0}_{1}'.format(parameter.lower(), channel)
    else:
        name = '{0}'.format(parameter.lower())

    webmap = None
    try:
        maps = cube.getMaps(plateifu=cube.plateifu, mode='local',
                            bintype=bintype, template=template)
        data = maps.getMap(parameter, channel=channel)

        # correct the stellar_sigma or emline_gsigma maps
        if parameter == 'stellar_sigma' or parameter == 'emline_gsigma':
            try:
                data = data.inst_sigma_correction()
            except MarvinError as e:
                pass
            else:
                name = 'Corrected {0}'.format(name)

    except Exception as e:
        mapmsg = 'Could not get map: {0}'.format(e)
    else:
        vals = data.value
        ivar = data.ivar
        mask = data.mask
        webmap = {'values': [it.tolist() for it in data.value],
                  'ivar': [it.tolist() for it in data.ivar] if data.ivar is not None else None,
                  'mask': [it.tolist() for it in data.mask] if data.mask is not None else None}
        mapmsg = "{0}: {1}-{2}".format(name, maps.bintype, maps.template)
    return webmap, mapmsg


def buildMapDict(cube, params, dapver, bintemp=None):
    ''' Build a list of dictionaries of maps

    params - list of string parameter names in form of category_channel

        NOT GENERALIZED
    '''
    # split the bintemp
    if bintemp:
        bintype, temp = bintemp.split('-', 1)
    else:
        bintype, temp = (None, None)

    mapdict = []
    params = params if isinstance(params, list) else [params]

    for param in params:
        param = str(param)
        try:
            parameter, channel = param.split(':')
        except ValueError as e:
            parameter, channel = (param, None)
        webmap, mapmsg = getWebMap(cube, parameter=parameter, channel=channel,
                                   bintype=bintype, template=temp)

        plotparams = datamodel[dapver].get_plot_params(prop=parameter)
        mask = Maskbit('MANGA_DAPPIXMASK')
        baddata_labels = [it for it in plotparams['bitmasks'] if it != 'NOCOV']
        baddata_bits = {it.lower(): int(mask.labels_to_bits(it)[0]) for it in baddata_labels}
        plotparams['bits'] = {'nocov': int(mask.labels_to_bits('NOCOV')[0]),
                              'badData': baddata_bits}
        mapdict.append({'data': webmap, 'msg': mapmsg, 'plotparams': plotparams})

    anybad = [m['data'] is None for m in mapdict]
    if any(anybad):
        bad_params = ', '.join([p for i, p in enumerate(params) if anybad[i]])
        raise MarvinError('Could not get map for: {0}.  Please select another.'.format(bad_params),
                          ignore_git=True)

    return mapdict


def make_nsa_dict(nsa, cols=None):
    ''' Make/rearrange the nsa dictionary of values '''

    # get columns
    if not cols:
        cols = [k for k in nsa.keys() if 'stokes' not in k]
        cols.sort()

    # make dictionary
    nsadict = {c: nsa[c] for c in cols}
    absmag = nsadict.get('elpetro_absmag', None)
    mtol = nsadict.get('elpetro_mtol', None)
    if absmag:
        nsadict.update({'elpetro_absmag_i': nsadict['elpetro_absmag'][5]})
    if mtol:
        nsadict.update({'elpetro_mtol_i': nsadict['elpetro_mtol'][5]})

    # check for logmass
    if 'elpetro_logmass' not in nsadict:
        if 'elpetro_mass' in nsadict:
            nsadict.update({'elpetro_logmass': np.log10(nsadict['elpetro_mass'])})
            cols.append('elpetro_logmass')

    cols.append('elpetro_mtol_i')
    cols.append('elpetro_absmag_i')
    cols.sort()

    return nsadict, cols


@functools.lru_cache(maxsize=128)
def get_nsa_dict(name, drpver, makenew=None):
    ''' Gets a NSA dictionary from a pickle or a query '''
    nsapath = os.environ.get('MANGA_SCRATCH_DIR', None)
    if nsapath and os.path.isdir(nsapath):
        nsapath = nsapath
    else:
        nsapath = os.path.expanduser('~')

    nsaroot = os.path.join(nsapath, 'nsa_pickles')
    if not os.path.isdir(nsaroot):
        os.makedirs(nsaroot)

    picklename = '{0}.pickle'.format(name)
    nsapickle_file = os.path.join(nsaroot, picklename)
    if os.path.isfile(nsapickle_file):
        nsadict = marvin_pickle.restore(nsapickle_file)
    else:
        # make from scratch from db
        marvindb = marvin.marvindb
        session = marvindb.session
        sampledb = marvindb.sampledb
        allnsa = session.query(sampledb.NSA, marvindb.datadb.Cube.plateifu).\
            join(sampledb.MangaTargetToNSA, sampledb.MangaTarget,
                 marvindb.datadb.Cube, marvindb.datadb.PipelineInfo,
                 marvindb.datadb.PipelineVersion, marvindb.datadb.IFUDesign).\
            filter(marvindb.datadb.PipelineVersion.version == drpver).options(
                FromCache(name)).options(*marvindb.cache_bits).all()
        nsadict = [(_db_row_to_dict(n[0], remove_columns=['pk', 'catalogue_pk']), n[1]) for n in allnsa]

        # write a new NSA pickle object
        if makenew:
            marvin_pickle.save(nsadict, path=nsapickle_file, overwrite=True)

    return nsadict


def remove_nans(datadict):
    ''' Removes objects with nan values from the NSA sample dictionary '''

    # collect total unique indices of nan objects
    allnans = np.array([])
    for key, vals in datadict.items():
        if key != 'plateifu':
            naninds = np.where(np.isnan(vals))[0]
            allnans = np.append(allnans, naninds)
    allnans = list(set(allnans.astype(np.int)))
    # delete those targets from all items in the dictionary
    for key, vals in datadict.items():
        datadict[key] = np.delete(np.asarray(vals), allnans).tolist()

    return datadict

vactarget_cache = {}


def create_vacdata(mangaid: str, release: str = None) -> dict:
    """ create a dictionary for VAC data for web display

    Parameters
    ----------
    mangaid : str
        The target mangaid 
    release : str
        The data release to use

    Returns
    -------
    dict
        A dictionary of VAC information
    """
    # hack to get vacs to switch based on release in the web
    # TODO - improve this
    release = release or marvin.config.release
    marvin.config.setRelease(release)
    from marvin.tools.vacs import VACs
    v = VACs()

    # get the VAC target info from the cache or filesystem
    if release in vactarget_cache and mangaid in vactarget_cache[release]:
        # may not hit since entire page is cached - see galaxy_get_cache
        dd = vactarget_cache[release][mangaid]
    else:
        dd = v.check_target(mangaid, lomem=True)
        # add to cache
        if release not in vactarget_cache:
            vactarget_cache[release] = {}
        vactarget_cache[release][mangaid] = dd

    vacdata=[]
    for i in v.list_vacs():
        if isinstance(v[i], dict):
            for k, j in v[i].items():
                link = j._rsync.url('', full=j._path)
                vacdata.append({'name': f'<a target="_blank" href="{j.url}">{j.display_name}: {k}</a>', 
                                'py': f'cube.vacs.{j.name}[{k}]', 'data': dd[i][k], 
                                'link': f'<a href="{link}">link</a>'})
        else:
            link = v[i]._rsync.url('', full=v[i]._path)
            vacdata.append({'name': f'<a target="_blank" href="{v[i].url}">{v[i].display_name}</a>',
                            'py': f'cube.vacs.{v[i].name}', 'data': dd[i], 
                            'link': f'<a href="{link}">link</a>'})
    return vacdata        


class Galaxy(BaseWebView):
    route_base = '/galaxy/'

    def __init__(self):
        ''' Initialize the route '''
        super(Galaxy, self).__init__('marvin-galaxy')
        self.galaxy = self.base.copy()
        self.galaxy['cube'] = None
        self.galaxy['image'] = ''
        self.galaxy['spectra'] = 'null'
        self.galaxy['maps'] = None
        self.galaxy['specmsg'] = None
        self.galaxy['mapmsg'] = None
        self.galaxy['toggleon'] = 'false'
        self.galaxy['nsamsg'] = None
        self.galaxy['hasnsa'] = False
        self.galaxy['nsachoices'] = {'1': {'y': 'z', 'x': 'elpetro_logmass', 'xtitle': 'Stellar Mass',
                                           'ytitle': 'Redshift', 'title': 'Redshift vs Stellar Mass'},
                                     '2': {'y': 'elpetro_absmag_g_r', 'x': 'elpetro_absmag_i', 'xtitle': 'AbsMag_i',
                                           'ytitle': 'Abs. g-r', 'title': 'Abs. g-r vs Abs. Mag i'}
                                     }
        # self.galaxy['nsachoices'] = {'1': {'y': 'z', 'x': 'sersic_mass', 'xtitle': 'Stellar Mass',
        #                                    'ytitle': 'Redshift', 'title': 'Redshift vs Stellar Mass'}}

        # cols = ['z', 'sersic_logmass', 'sersic_n', 'sersic_absmag', 'elpetro_mag_g_r', 'elpetro_th50_r']
        self.galaxy['nsaplotcols'] = ['z', 'elpetro_logmass', 'sersic_n', 'elpetro_absmag_i', 'elpetro_absmag_g_r',
                                      'elpetro_th50_r', 'elpetro_absmag_u_r', 'elpetro_absmag_i_z', 'elpetro_ba',
                                      'elpetro_phi', 'elpetro_mtol_i', 'elpetro_th90_r']
        self.galaxy['nsadict'] = {}

    def before_request(self, *args, **kwargs):
        ''' Do these things before a request to any route '''
        super(Galaxy, self).before_request(*args, **kwargs)
        self.reset_dict(self.galaxy, exclude=['nsachoices', 'nsaplotcols', 'cube'])

    def index(self):
        ''' Main galaxy page '''
        self.galaxy['error'] = 'Not all there are you...  Try adding a plate-IFU or manga-ID to the end of the address.'
        return render_template("galaxy.html", **self.galaxy)

    @cache.memoize(timeout=300)
    def get(self, galid):
        ''' Retrieve info for a given cube '''

        # determine type of galid
        args = av.manual_parse(self, request, use_params='galaxy')
        self.galaxy['id'] = args['galid']
        self.galaxy['latest_dr'] = self._release.lower() if 'DR' in self._release else marvin.config._get_latest_release(dr_only=True).lower()
        idtype = parseIdentifier(galid)
        if idtype in ['plateifu', 'mangaid']:
            # set plateifu or mangaid
            self.galaxy['idtype'] = idtype
            galaxyid = {self.galaxy['idtype']: galid, 'release': self._release}
            self.galaxy['vacdata'] = []

            # Get cube
            try:
                cube = Cube(**galaxyid)
            except MarvinError as e:
                self.galaxy['cube'] = None
                errmsg = 'MarvinError: {0}'.format(e)

                # check target status and fine-tune the error message
                if idtype == 'mangaid':
                    status = target_status(galid, drpver=self._drpver)
                    if status == 'not yet observed':
                        errmsg = '{0} is a valid target but has not yet been observed'.format(galid)
                    elif status == 'not valid target':
                        errmsg = '{0} is not valid MaNGA target.  Check your syntax'.format(galid)

                self.galaxy['error'] = errmsg
                return render_template("galaxy.html", **self.galaxy)
            else:
                dm = datamodel[self._dapver]
                self.galaxy['cube'] = cube
                g.cube = cube
                self.galaxy['daplink'] = getDapRedux(release=self._release)
                # get SAS url links to cube, rss, maps, image
                if Path:
                    sdss_path = Path(release=self._release)
                    self.galaxy['image'] = cube.getImage().url
                    cubelink = sdss_path.url('', full=cube._getFullPath())
                    rsslink = sdss_path.url('mangarss', drpver=cube._drpver, plate=cube.plate, ifu=cube.ifu, wave='LOG')
                    daptype = "{0}-{1}".format(dm.default_bintype, dm.default_template)
                    maplink = getDefaultMapPath(release=self._release, plate=cube.plate, ifu=cube.ifu, daptype=daptype, mode='MAPS')
                    mclink = getDefaultMapPath(release=self._release, plate=cube.plate, ifu=cube.ifu, daptype=daptype, mode='LOGCUBE')
                    self.galaxy['links'] = {'cube': cubelink, 'rss': rsslink, 'map': maplink, 'mc': mclink}
                else:
                    self.galaxy['image'] = cube.data.image

                infile = get_manga_image(cube, local=True)
                current_session['imagefile'] = infile
                current_session['wcs'] = cube.wcs.to_header_string()
                current_session['cube_shape'] = cube.wcs.array_shape[-1]

            # Get the initial spectrum
            if cube:
                dm = datamodel[self._dapver]
                daplist = [p.full(web=True) for p in dm.properties]
                dapdefaults = dm.get_default_mapset()
                self.galaxy['cube'] = cube
                self.galaxy['vacdata'] = create_vacdata(cube.mangaid, release=self._release)             
                self.galaxy['toggleon'] = current_session.get('toggleon', 'false')
                self.galaxy['cubehdr'] = cube.header
                self.galaxy['quality'] = ('DRP3QUAL', cube.quality_flag.mask, cube.quality_flag.labels)
                self.galaxy['mngtarget'] = {'bits': [it.mask for it in cube.target_flags if it.mask != 0],
                                            'labels': [it.labels for it in cube.target_flags if len(it.labels) > 0],
                                            'names': [''.join(('MNGTARG', it.name[-1])) for it in cube.target_flags if it.mask != 0]}

                # check if the cube has any nsa data
                # catch general exceptions
                try:
                    hasnsa = cube.nsa is not None
                except Exception:
                    hasnsa = False

                # make the nsa dictionary
                self.galaxy['hasnsa'] = hasnsa
                if hasnsa:
                    cols = self.galaxy.get('nsaplotcols')
                    nsadict, nsacols = make_nsa_dict(cube.nsa)
                    nsatmp = [nsacols.pop(nsacols.index(i)) for i in cols if i in nsacols]
                    nsatmp.extend(nsacols)
                    self.galaxy['nsacols'] = nsatmp
                    self.galaxy['nsadict'] = nsadict

                self.galaxy['dapmaps'] = daplist
                self.galaxy['dapmapselect'] = current_session.get('selected_dapmaps', dapdefaults)
                dm = datamodel[self._dapver]
                self.galaxy['dapbintemps'] = dm.get_bintemps(db_only=True)
                if 'bintemp' not in current_session or current_session['bintemp'] not in self.galaxy['dapbintemps']:
                    current_session['bintemp'] = '{0}-{1}'.format(dm.get_bintype(), dm.get_template())

                # get default map quality
                try:
                    maps = cube.getMaps(plateifu=cube.plateifu, mode='local', bintype=current_session['bintemp'].split('-')[0])
                except MarvinError:
                    mapqual = ('DAPQUAL', [], [])
                else:
                    mapqual = ('DAPQUAL', maps.quality_flag.mask, maps.quality_flag.labels)
                self.galaxy['mapquality'] = mapqual

                # TODO - make this general - see also search.py for querystr
                self.galaxy['cubestr'] = ("<html><samp>from marvin.tools.cube import Cube<br>cube = \
                    Cube(plateifu='{0}')<br># access the header<br>cube.header<br># get NSA data<br>\
                    cube.nsa<br></samp></html>".format(cube.plateifu))

                self.galaxy['spaxelstr'] = ("<html><samp>from marvin.tools.cube import Cube<br>cube = \
                    Cube(plateifu='{0}')<br># get a spaxel by slicing cube[i,j]<br>spaxel=cube[16, 16]<br>flux = \
                    spaxel.flux<br>wave = flux.wavelength<br>ivar = flux.ivar<br>mask = \
                    flux.mask<br>flux.plot()<br></samp></html>".format(cube.plateifu))

                self.galaxy['mapstr'] = ("<html><samp>from marvin.tools.maps import Maps<br>maps = \
                    Maps(plateifu='{0}')<br>print(maps)<br># get an emission \
                    line map<br>haflux = maps.emline_gflux_ha_6564<br>values = \
                    haflux.value<br>ivar = haflux.ivar<br>mask = haflux.mask<br>haflux.plot()<br>\
                    </samp></html>".format(cube.plateifu))

                self.galaxy['vactoolstr'] = ("<html><samp>from marvin.tools.vacs import VACs<br>\
                    vacs = VACs()</samp></html>")

                self.galaxy['vactargstr'] = ("<html><samp>from marvin.tools.cube import Cube<br>\
                    cube = Cube(mangaid='{0}')<br>cube.vacs</samp></html>".format(cube.mangaid))

        else:
            self.galaxy['error'] = 'Error: Galaxy ID {0} must either be a Plate-IFU, or MaNGA-Id designation.'.format(galid)
            return render_template("galaxy.html", **self.galaxy)

        return render_template("galaxy.html", **self.galaxy)

    @route('/initdynamic/', methods=['POST'], endpoint='initdynamic')
    def initDynamic(self):
        ''' Route to run when the dynamic toggle is initialized
            This creates the web spectrum and dap heatmaps
        '''

        # get the form parameters
        args = av.manual_parse(self, request, use_params='galaxy', required='plateifu')

        # datamodel
        dm = datamodel[self._dapver]

        # turning toggle on
        nowebsession = marvin.config._custom_config.get('no_web_session', None)
        if not nowebsession:
            current_session['toggleon'] = args.get('toggleon')

        # get the cube
        cubeinputs = {'plateifu': args.get('plateifu'), 'release': self._release}
        cube = Cube(**cubeinputs)
        output = {'specstatus': -1, 'mapstatus': -1}

        # get web spectrum
        webspec, specmsg, badspots = getWebSpectrum(cube, cube.ra, cube.dec, byradec=True)
        daplist = [p.full(web=True) for p in dm.properties]
        dapdefaults = dm.get_default_mapset()

        # select any DAP maps and bin-template from the session
        selected_bintemp = current_session.get('bintemp', None)
        selected_maps = current_session.get('selected_dapmaps', dapdefaults)

        # check for correct bintemp
        if selected_bintemp and selected_bintemp not in dm.get_bintemps(db_only=True):
            selected_bintemp = '{0}-{1}'.format(dm.get_bintype(), dm.get_template())

        # check that selected bintemp is available for target
        hasbin = selected_bintemp.split('-')[0] in cube.get_available_bintypes() if selected_bintemp else None

        # build the uber map dictionary
        try:
            mapdict = buildMapDict(cube, selected_maps, self._dapver, bintemp=selected_bintemp)
            mapmsg = None
        except Exception as e:
            mapdict = [{'data': None, 'msg': 'Error', 'plotparams': None} for m in dapdefaults]
            if hasbin:
                mapmsg = 'Error getting maps: {0}'.format(e)
            else:
                mapmsg = 'No maps available for selected bintype {0}. Try a different one.'.format(selected_bintemp)
        else:
            output['mapstatus'] = 1

        if not webspec:
            output['error'] = 'Error: {0}'.format(specmsg)
        else:
            output['specstatus'] = 1

        output['image'] = get_manga_image(cube)
        output['spectra'] = webspec
        output['specmsg'] = specmsg
        output['maps'] = mapdict
        output['mapmsg'] = mapmsg
        output['dapmaps'] = daplist
        output['dapmapselect'] = selected_maps
        output['daplines'] = dm.get_channels('emline_gflux', formatted=True)
        output['badspots'] = badspots

        output['dapbintemps'] = dm.get_bintemps(db_only=True)
        if 'bintemp' not in current_session:
            current_session['bintemp'] = '{0}-{1}'.format(dm.get_bintype(), dm.get_template())

        # try to jsonify the result
        try:
            jsonout = jsonify(result=output)
        except Exception as e:
            jsonout = jsonify(result={'specstatus': -1, 'mapstatus': -1, 'error': '{0}'.format(e)})

        return jsonout

    @route('/getspaxel/', methods=['POST'], endpoint='getspaxel')
    @cache.memoize()
    def getSpaxel(self):
        args = av.manual_parse(self, request, use_params='galaxy', required=['plateifu', 'type'], makemulti=True)
        cubeinputs = {'plateifu': args.get('plateifu'), 'release': self._release}
        maptype = args.get('type', None)

        if maptype == 'optical':
            # for now, do this, but TODO - general processRequest to handle lists and not lists
            try:
                mousecoords = args.getlist('mousecoords[]', type=float)
            except Exception as e:
                mousecoords = None

            if mousecoords:
                pixshape = (args.get('imwidth', type=int), args.get('imheight', type=int))
                if (mousecoords[0] < 0 or mousecoords[0] > pixshape[0]) or (mousecoords[1] < 0 or mousecoords[1] > pixshape[1]):
                    output = {'specmsg': 'Error: requested pixel coords are outside the image range.', 'status': -1}
                    self.galaxy['error'] = output['specmsg']
                else:
                    cube = Cube(**cubeinputs)

                    # TODO - generalize image file sas_url to filesystem switch, maybe in sdss_access
                    infile = get_manga_image(cube, local=True)
                    current_session['imagefile'] = infile
                    #infile = os.path.join(os.getenv('MANGA_SPECTRO_REDUX'), args.get('image').split('redux/')[1])
                    arrcoords = convertImgCoords(mousecoords, infile, to_radec=True)

                    webspec, specmsg, badspots = getWebSpectrum(cube, arrcoords[0], arrcoords[1], byradec=True)
                    if not webspec:
                        self.galaxy['error'] = 'Error: {0}'.format(specmsg)
                        status = -1
                    else:
                        status = 1
                    msg = 'gettin some spaxel at RA/Dec {0}'.format(arrcoords)
                    output = {'message': msg, 'specmsg': specmsg, 'spectra': webspec, 'status': status, 'badspots': badspots}
            else:
                output = {'specmsg': 'Error getting mouse coords', 'status': -1}
                self.galaxy['error'] = output['specmsg']
        elif maptype == 'heatmap':
            # grab spectrum based on (x, y) coordinates
            x = args.get('x', None, type=int)
            y = args.get('y', None, type=int)
            if all([x, y]):
                cube = Cube(**cubeinputs)
                webspec, specmsg, badspots = getWebSpectrum(cube, x, y, xyorig='lower')
                msg = 'gettin some spaxel with (x={0}, y={1})'.format(x, y)
                if not webspec:
                    self.galaxy['error'] = 'Error: {0}'.format(specmsg)
                    status = -1
                else:
                    status = 1
                output = {'message': msg, 'specmsg': specmsg, 'spectra': webspec, 'status': status, 'badspots': badspots}
            else:
                output = {'specmsg': 'Error: X or Y not specified for map', 'status': -1}
                self.galaxy['error'] = output['specmsg']
        else:
            output = {'specmsg': 'Error: No maptype specified in request', 'status': -1}
            self.galaxy['error'] = output['specmsg']

        return jsonify(result=output)

    @route('/updatemaps/', methods=['POST'], endpoint='updatemaps')
    @cache.memoize()
    def updateMaps(self):
        args = av.manual_parse(self, request, use_params='galaxy', required=['plateifu', 'bintemp', 'params[]'], makemulti=True)
        #self._drpver, self._dapver, self._release = parseSession()
        cubeinputs = {'plateifu': args.get('plateifu'), 'release': self._release}
        params = args.getlist('params[]', type=str)
        bintemp = args.get('bintemp', None, type=str)

        # update the session variables
        current_session['bintemp'] = bintemp
        current_session['selected_dapmaps'] = params

        # get cube (self.galaxy['cube'] does not work)
        try:
            cube = Cube(**cubeinputs)
        except Exception as e:
            cube = None
        # Try to make the web maps
        if not cube:
            output = {'mapmsg': 'No cube found', 'maps': None, 'status': -1}
        elif not params:
            output = {'mapmsg': 'No parameters selected', 'maps': None, 'status': -1}
        else:
            try:
                mapdict = buildMapDict(cube, params, self._dapver, bintemp=bintemp)
            except Exception as e:
                output = {'mapmsg': str(e), 'status': -1, 'maps': None}
            else:
                output = {'mapmsg': None, 'status': 1, 'maps': mapdict}
        return jsonify(result=output)

    @route('/initnsaplot/', methods=['POST'], endpoint='initnsaplot')
    @cache.memoize()
    def init_nsaplot(self):
        args = av.manual_parse(self, request, use_params='galaxy', required='plateifu')
        #self._drpver, self._dapver, self._release = parseSession()
        cubeinputs = {'plateifu': args.get('plateifu'), 'release': self._release}

        # get the default nsa choices
        nsachoices = self.galaxy.get('nsachoices', None)
        if not nsachoices:
            nsachoices = {'1': {'y': 'z', 'x': 'elpetro_logmass', 'xtitle': 'Stellar Mass',
                                'ytitle': 'Redshift', 'title': 'Redshift vs Stellar Mass'},
                          '2': {'y': 'elpetro_absmag_g_r', 'x': 'elpetro_absmag_i', 'xtitle': 'AbsMag_i',
                                'ytitle': 'Abs. g-r', 'title': 'Abs. g-r vs Abs. Mag i'}}

        # get cube (self.galaxy['cube'] does not work)
        try:
            cube = Cube(**cubeinputs)
        except Exception as e:
            cube = None

        # get some nsa params
        if not cube:
            output = {'nsamsg': 'No cube found', 'nsa': None, 'status': -1}
        else:
            # get the galaxy nsa parameters
            cols = self.galaxy.get('nsaplotcols')
            try:
                nsadict, nsacols = make_nsa_dict(cube.nsa)
                nsa = {args.get('plateifu'): nsadict}
            except Exception as e:
                output = {'nsamsg': str(e), 'status': -1, 'nsa': None}
            else:
                # get the sample nsa parameters
                try:
                    nsacache = 'nsa_{0}'.format(self._release.lower().replace('-', ''))
                    nsadict = get_nsa_dict(nsacache, self._drpver)
                except Exception as e:
                    output = {'nsamsg': 'Failed to retrieve sample NSA: {0}'.format(e), 'status': -1, 'nsa': nsa, 'nsachoices': nsachoices}
                else:
                    #nsadict = [(_db_row_to_dict(n[0], remove_columns=['pk', 'catalogue_pk']), n[1]) for n in allnsa]
                    nsasamp = {c: [n[0][c.split('_i')[0]][5] if 'absmag_i' in c or 'mtol_i' in c else n[0][c] for n in nsadict] for c in cols}
                    nsasamp['plateifu'] = [n[1] for n in nsadict]
                    nsasamp = remove_nans(nsasamp)
                    nsa['sample'] = nsasamp
                    output = {'nsamsg': None, 'status': 1, 'nsa': nsa, 'nsachoices': nsachoices, 'nsaplotcols': cols}
        return jsonify(result=output)


def get_spaxel_cache(*args, **kwargs):
    ''' Function used to generate the route cache key

    Cache key when using cache.memoize or cache.cached decorator.
    memoize remembers input methods arguments; cached does not.

    Parameters:
        args (list):
            a list of the fx/method route call and object instance (self)
        kwargs (dict):
            a dictonary of arguments passed into the method
    Returns:
        A string used for the cache key lookup
    '''
    # get the method and self instance
    fxn, inst = args

    # parse the form request to extract any parameters
    reqargs = av.manual_parse(inst, request, use_params='galaxy', required=['plateifu', 'type'], makemulti=True)

    maptype = reqargs.get('type', None)
    if maptype == 'optical':
        mousecoords = reqargs.getlist('mousecoords[]', type=float)
        infile = current_session['imagefile']
        radec = convertImgCoords(mousecoords, infile, to_radec=True)
        ww = WCS(current_session['wcs'])
        cube_shape = [current_session['cube_shape']] * 2
        y, x = convertCoords(radec, wcs=ww, shape=cube_shape,
                             mode='sky', xyorig='lower').T
        y = y[0]
        x = x[0]
    elif maptype == 'heatmap':
        x = reqargs.get('x', None, type=int)
        y = reqargs.get('y', None, type=int)

    plateifu = reqargs.get('plateifu').replace('-', '_')
    release = inst._release.lower().replace('-', '')

    # create unique cache key name
    key = 'getspaxel_{0}_{1}_{2}_{3}'.format(
        release, plateifu, x, y)

    return key


def get_nsa_cache(*args, **kwargs):
    ''' Function used to generate the route cache key

    Cache key when using cache.memoize or cache.cached decorator.
    memoize remembers input methods arguments; cached does not.

    Parameters:
        args (list):
            a list of the fx/method route call and object instance (self)
        kwargs (dict):
            a dictonary of arguments passed into the method
    Returns:
        A string used for the cache key lookup
    '''
    # get the method and self instance
    fxn, inst = args

    # parse the form request to extract any parameters
    reqargs = av.manual_parse(inst, request, use_params='galaxy', required='plateifu')

    plateifu = reqargs.get('plateifu').replace('-', '_')
    release = inst._release.lower().replace('-', '')

    # create unique cache key name
    key = 'getnsa_{0}_{1}'.format(release, plateifu)

    return key


def update_maps_cache(*args, **kwargs):
    ''' Function used to generate the route cache key

    Cache key when using cache.memoize or cache.cached decorator.
    memoize remembers input methods arguments; cached does not.

    Parameters:
        args (list):
            a list of the fx/method route call and object instance (self)
        kwargs (dict):
            a dictonary of arguments passed into the method
    Returns:
        A string used for the cache key lookup
    '''
    # get the method and self instance
    fxn, inst = args

    # parse the form request to extract any parameters
    reqargs = av.manual_parse(inst, request, use_params='galaxy', required=[
                              'plateifu', 'bintemp', 'params[]'], makemulti=True)

    plateifu = reqargs.get('plateifu').replace('-', '_')
    release = inst._release.lower().replace('-', '')

    bintemp = reqargs.get('bintemp', None, type=str).replace('-', '_')
    params = reqargs.getlist('params[]', type=str)
    params.sort()
    param_string = '_'.join(params).replace(':', '_')

    # create unique cache key name
    key = 'updatemaps_{0}_{1}_{2}_{3}'.format(release, plateifu, bintemp, param_string)

    return key


def galaxy_get_cache(*args, **kwargs):
    ''' Function used to generate the route cache key

    Cache key when using cache.memoize or cache.cached decorator.
    memoize remembers input methods arguments; cached does not.

    Parameters:
        args (list):
            a list of the fx/method route call and object instance (self)
        kwargs (dict):
            a dictonary of arguments passed into the method
    Returns:
        A string used for the cache key lookup
    '''

    # get the method and self instance
    fxn, inst = args

    # parse the form request to extract any parameters
    reqargs = av.manual_parse(inst, request, use_params='galaxy')

    galid = reqargs.get('galid').replace('-', '_')
    release = inst._release.lower().replace('-', '')

    # create unique cache key name
    key = 'getpage_{0}_{1}'.format(release, galid)
    # append if logged in
    if inst.galaxy['loggedin']:
        key = '{0}_loggedin'.format(key)

    return key


Galaxy.getSpaxel.make_cache_key = get_spaxel_cache
Galaxy.init_nsaplot.make_cache_key = get_nsa_cache
Galaxy.updateMaps.make_cache_key = update_maps_cache
Galaxy.get.make_cache_key = galaxy_get_cache

Galaxy.register(galaxy)
